from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

__all__ = ["Op"]

from bewer.alignment.op_type import OpType

if TYPE_CHECKING:
    from bewer.alignment.alignment import Alignment


class Op:
    """Class representing an operation with its type.

    Attributes:
        type (OpType): The type of operation.
        ref (str | None): The reference string involved in the operation.
        hyp (str | None): The hypothesis string involved in the operation.
        ref_token_idx (int | None): The index of the token in the reference text (for word-level alignments).
        hyp_token_idx (int | None): The index of the token in the hypothesis text (for word-level alignments).
        ref_span (slice | None): The index span in the reference text.
        hyp_span (slice | None): The index span in the hypothesis text.
        hyp_left_partial (bool): Whether the hypothesis token is a partial word, cropped on the left.
        hyp_right_partial (bool): Whether the hypothesis token is a partial word, cropped on the right.
        ref_left_partial (bool): Whether the reference token is a partial word, cropped on the left.
        ref_right_partial (bool): Whether the reference token is a partial word, cropped on the right.
    """

    def __init__(
        self,
        type: OpType,
        ref: str | None = None,
        hyp: str | None = None,
        ref_token_idx: int | None = None,
        hyp_token_idx: int | None = None,
        ref_span: slice | None = None,
        hyp_span: slice | None = None,
        hyp_left_partial: bool = False,
        hyp_right_partial: bool = False,
        ref_left_partial: bool = False,
        ref_right_partial: bool = False,
        src: "Alignment" | None = None,
    ):
        """Initialize Op.

        Args:
            type: Operation type (MATCH, INSERT, DELETE, SUBSTITUTE).
            ref: Reference string.
            hyp: Hypothesis string.
            ref_token_idx: Reference token index.
            hyp_token_idx: Hypothesis token index.
            ref_span: Reference span.
            hyp_span: Hypothesis span.
            hyp_left_partial: Whether hypothesis is left-partial.
            hyp_right_partial: Whether hypothesis is right-partial.
            ref_left_partial: Whether reference is left-partial.
            ref_right_partial: Whether reference is right-partial.
            src: Parent Alignment object. Can be set later via set_source().
        """
        self.type = type
        self.ref = ref
        self.hyp = hyp
        self.ref_token_idx = ref_token_idx
        self.hyp_token_idx = hyp_token_idx
        self._ref_span = ref_span
        self._hyp_span = hyp_span
        self.hyp_left_partial = hyp_left_partial
        self.hyp_right_partial = hyp_right_partial
        self.ref_left_partial = ref_left_partial
        self.ref_right_partial = ref_right_partial

        self._src = None
        if src is not None:
            self.set_source(src)

        if self.type == OpType.MATCH:
            if self.ref is None or self.hyp is None:
                raise ValueError("MATCH operation must have non-empty ref or hyp.")
        elif self.type == OpType.INSERT:
            if self.hyp is None or self.ref is not None:
                raise ValueError("INSERT operation must have non-empty hyp and empty ref.")
        elif self.type == OpType.DELETE:
            if self.hyp is not None or self.ref is None:
                raise ValueError("DELETE operation must have non-empty ref and empty hyp.")
        elif self.type == OpType.SUBSTITUTE:
            if self.ref is None or self.hyp is None:
                raise ValueError("SUBSTITUTE operation must have both ref and hyp.")

    @property
    def _repr_hyp(self) -> str | None:
        """Return the hypothesis with partial markers if applicable."""
        if self.hyp is None:
            return None
        return f'{"-" if self.hyp_left_partial else ""}"{self.hyp}"{"-" if self.hyp_right_partial else ""}'

    @property
    def _repr_ref(self) -> str | None:
        """Return the reference with partial markers if applicable."""
        if self.ref is None:
            return None
        return f'{"-" if self.ref_left_partial else ""}"{self.ref}"{"-" if self.ref_right_partial else ""}'

    @property
    def src(self) -> "Alignment" | None:
        """Get the parent Alignment object."""
        return self._src

    def set_source(self, src: "Alignment") -> None:
        """Set the parent Alignment object.

        Args:
            src: The parent Alignment object.

        Raises:
            ValueError: If source is already set.
        """
        if self._src is not None:
            raise ValueError("Source already set for Op")
        self._src = src

    @cached_property
    def ref_span(self) -> slice | None:
        """Get the reference span for an operation."""
        if self._ref_span is not None:
            return self._ref_span
        if self.ref_token_idx is not None:
            if self._src is None or self._src.src is None:
                return None
            return self._src.src.ref.tokens[self.ref_token_idx].slice
        return None

    @cached_property
    def hyp_span(self) -> slice | None:
        """Get the hypothesis span for an operation."""
        if self._hyp_span is not None:
            return self._hyp_span
        if self.hyp_token_idx is not None:
            if self._src is None or self._src.src is None:
                return None
            return self._src.src.hyp.tokens[self.hyp_token_idx].slice
        return None

    def to_dict(self) -> dict:
        """Convert the Op instance to a dictionary."""
        return {
            "type": self.type.name,
            "ref": self.ref,
            "hyp": self.hyp,
            "ref_token_idx": self.ref_token_idx,
            "hyp_token_idx": self.hyp_token_idx,
            "ref_span": (self.ref_span.start, self.ref_span.stop) if self.ref_span else None,
            "hyp_span": (self.hyp_span.start, self.hyp_span.stop) if self.hyp_span else None,
            "hyp_left_partial": self.hyp_left_partial,
            "hyp_right_partial": self.hyp_right_partial,
            "ref_left_partial": self.ref_left_partial,
            "ref_right_partial": self.ref_right_partial,
        }

    def __repr__(self) -> str:
        if self.type == OpType.DELETE:
            return f"Op({self.type.name}: {self._repr_ref})"
        if self.type == OpType.INSERT:
            return f"Op({self.type.name}: {self._repr_hyp})"
        if self.type == OpType.SUBSTITUTE:
            return f"Op({self.type.name}: {self._repr_hyp} -> {self._repr_ref})"
        return f"Op({self.type.name}: {self._repr_hyp} == {self._repr_ref})"



from __future__ import annotations

import json
from collections import Counter
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, Optional, Union

from bewer.alignment.op_type import OpType
from bewer.reporting.html.alignment import generate_alignment_html_lines as _generate_alignment_html_lines
from bewer.reporting.html.color_schemes import HTMLDefaultAlignmentColors
from bewer.reporting.python.alignment import DefaultColorScheme, display_basic_aligned

if TYPE_CHECKING:
    from bewer.alignment.op import Op
    from bewer.core.example import Example
    from bewer.reporting.html.color_schemes import HTMLAlignmentColors
    from bewer.reporting.python.alignment import ColorScheme

__all__ = ["Alignment"]


class Alignment(tuple["Op", ...]):
    """
    Immutable sequence of operations with additional methods for processing.

    Attributes:
        num_matches (int): Number of match operations.
        num_substitutions (int): Number of substitution operations.
        num_insertions (int): Number of insertion operations.
        num_deletions (int): Number of deletion operations.
    """

    def __new__(cls, iterable: Iterable["Op"] = (), src: Optional["Example"] = None) -> "Alignment":
        return super().__new__(cls, iterable)

    def __init__(self, iterable: Iterable["Op"] = (), src: Optional["Example"] = None) -> None:
        """Initialize Alignment.

        Args:
            iterable: Iterable of Op objects.
            src: Parent Example object. Can be set later via set_source().
        """
        self._op_counts = Counter()
        for op in self:
            self._op_counts[op.type] += 1
            if op.src is None:
                op.set_source(self)

        self._src = None
        if src is not None:
            self.set_source(src)

    @property
    def num_matches(self) -> int:
        """Get the number of match operations."""
        return self._op_counts[OpType.MATCH]

    @property
    def num_substitutions(self) -> int:
        """Get the number of substitution operations."""
        return self._op_counts[OpType.SUBSTITUTE]

    @property
    def num_insertions(self) -> int:
        """Get the number of insertion operations."""
        return self._op_counts[OpType.INSERT]

    @property
    def num_deletions(self) -> int:
        """Get the number of deletion operations."""
        return self._op_counts[OpType.DELETE]

    @property
    def num_edits(self) -> int:
        """Get the total number of edit operations (substitutions, insertions, deletions)."""
        return self.num_substitutions + self.num_insertions + self.num_deletions

    @property
    def num_ops(self) -> int:
        """Get the total number of operations."""
        return len(self)

    @cached_property
    def _start_index_mapping(self) -> dict[int, int]:
        """Create a mapping from character start index to token index for quick lookup."""
        mapping = {}
        for i, op in enumerate(self):
            if op.ref_span is not None:
                mapping[op.ref_span.start] = i
        return mapping

    @cached_property
    def _end_index_mapping(self) -> dict[int, int]:
        """Create a mapping from character end index to token index for quick lookup."""
        mapping = {}
        for i, op in enumerate(self):
            if op.ref_span is not None:
                mapping[op.ref_span.stop] = i
        return mapping

    @cached_property
    def ref_index_mapping(self) -> dict[int, int]:
        """Create a mapping from reference token index to operation index for quick lookup."""
        mapping = {}
        for i, op in enumerate(self):
            if op.ref_token_idx is not None:
                mapping[op.ref_token_idx] = i
        return mapping

    def ops_from_ref_index(self, start: int, stop: Optional[int] = None) -> Alignment:
        """Get the ops that correspond to the given reference token index or slice.

        Args:
            start (int): The start reference token index.
            stop (Optional[int]): The stop reference token index.

        Returns:
            Alignment: The ops that correspond to the given reference token span.

        Raises:
            ValueError: If start or stop indices are not found in ref_index_mapping, or if stop index is less than
                start index.
        """
        if start not in self.ref_index_mapping:
            raise ValueError(f"Start index {start} not found in reference index mapping.")
        if stop is not None and stop not in self.ref_index_mapping:
            raise ValueError(f"Stop index {stop} not found in reference index mapping.")
        if stop is not None:
            if stop < start:
                raise ValueError("Stop index must be greater than or equal to start index.")
            return self[self.ref_index_mapping[start] : self.ref_index_mapping[stop] + 1]
        return self[self.ref_index_mapping[start] : self.ref_index_mapping[start] + 1]

    def start_index_to_op(self, char_index: int) -> Optional[Op]:
        """Get the op that starts at the given character index.

        Args:
            char_index (int): The character index to look up.

        Returns:
            Optional[Op]: The op that starts at the given character index, or None if not found.
        """
        op_index = self._start_index_mapping.get(char_index, None)
        if op_index is not None:
            return self[op_index]
        return None

    def end_index_to_op(self, char_index: int) -> Optional[Op]:
        """Get the op that ends at the given character index.

        Args:
            char_index (int): The character index to look up.

        Returns:
            Optional[Op]: The op that ends at the given character index, or None if not found.
        """
        op_index = self._end_index_mapping.get(char_index, None)
        if op_index is not None:
            return self[op_index]
        return None

    @property
    def src(self) -> Optional["Example"]:
        """Get the parent Example object."""
        return self._src

    def set_source(self, src: "Example") -> None:
        """Set the parent Example object.

        Args:
            src: The parent Example object.

        Raises:
            ValueError: If source is already set.
        """
        if self._src is not None:
            raise ValueError("Source already set for Alignment")
        self._src = src

    def to_dicts(self) -> list[dict]:
        """Dump the alignment to a list of dictionaries.

        Returns:
            list[dict]: List of operations as dictionaries.
        """
        return [op.to_dict() for op in self]

    def to_json(self, path: str | None = None, allow_overwrite: bool = False) -> str:
        """Dump the alignment to a JSON string.

        Args:
            path (str | None): If provided, write the JSON string to this file.
            allow_overwrite (bool): If True, overwrite the file if it exists.
        Returns:
            str: JSON string representing the alignment.
        """
        json_str = json.dumps(self.to_dicts(), indent=2)
        if path is not None:
            path = Path(path)
            if path.is_dir():
                raise ValueError("Provided path is a directory, expected a file path.")
            if path.exists() and not allow_overwrite:
                raise FileExistsError(f"File {path} already exists.")
            if not path.parent.exists():
                path.parent.mkdir(parents=True, exist_ok=True)
            with open(path, "w") as f:
                f.write(json_str)
        return json_str

    def display(
        self,
        max_line_length: int | float = 0.5,
        color_scheme: ColorScheme = DefaultColorScheme,
    ) -> None:
        """Display the alignment in the console.

        Args:
            max_line_length (int | float): Maximum line length for display. If a float between 0 and 1,
                interpreted as a fraction of the terminal width.
            color_scheme (ColorScheme): Color scheme for display.
        """
        title = None if self._src is None else f"   Example {self._src.index}"
        display_basic_aligned(self, max_line_length=max_line_length, title=title, color_scheme=color_scheme)

    def _to_html_lines(
        self,
        color_scheme: type[HTMLAlignmentColors] = HTMLDefaultAlignmentColors,
    ) -> list[tuple[str, str]]:
        """Render the alignment as an HTML string.

        Args:
            color_scheme (type[HTMLAlignmentColors]): Color scheme for display.

        Returns:
            list[tuple[str, str]]: List of tuples containing HTML strings representing the alignment visualization.
        """
        return _generate_alignment_html_lines(self, color_scheme=color_scheme)

    def __getitem__(self, index: int | slice) -> Union[Op, "Alignment"]:
        if isinstance(index, slice):
            return Alignment(super().__getitem__(index))
        return super().__getitem__(index)

    def __repr__(self):
        ops = self[:60]
        ops_str = ",\n ".join([repr(op) for op in ops])
        if len(self) > 60:
            ops_str += ",\n ..."
        return f"Alignment([\n {ops_str}]\n)"



"""Configurable labels and tooltips for alignment display in HTML reports."""

__all__ = ["HTMLAlignmentLabels"]


class HTMLAlignmentLabels:
    """Configurable labels and tooltips for alignment display in HTML reports.

    Subclass and override any attribute to customize the labels shown in the report.
    Follows the same subclassing pattern as HTMLAlignmentColors.
    """

    # Line indicators
    REF = "Ref."
    HYP = "Hyp."

    # Legend labels
    MATCH = "Match"
    SUBSTITUTION = "Substitution"
    INSERTION = "Insertion"
    DELETION = "Deletion"
    PADDING = "Padding"
    KEYWORD = "Keyword"

    # Legend tooltips (None = no tooltip rendered)
    MATCH_TOOLTIP: str | None = None
    SUBSTITUTION_TOOLTIP: str | None = None
    INSERTION_TOOLTIP: str | None = None
    DELETION_TOOLTIP: str | None = None
    PADDING_TOOLTIP: str | None = None
    KEYWORD_TOOLTIP: str | None = None



from bewer.metrics import corti_legacy_metrics  # noqa (for backward compatibility)
from bewer.metrics.base import (
    METRIC_REGISTRY,
    ExampleMetric,
    Metric,
    MetricParams,
    list_registered_metrics,
    metric_value,
)
from bewer.metrics.cer import CER
from bewer.metrics.error_align import ErrorAlign
from bewer.metrics.kwer import KWER
from bewer.metrics.levenshtein import Levenshtein
from bewer.metrics.mtr import MTR
from bewer.metrics.summary import DatasetSummary

# Metric implementations
from bewer.metrics.wer import WER

__all__ = [
    "METRIC_REGISTRY",
    "list_registered_metrics",
    "Metric",
    "ExampleMetric",
    "MetricParams",
    "metric_value",
    "WER",
    "CER",
    "KWER",
    "MTR",
    "Levenshtein",
    "ErrorAlign",
    "DatasetSummary",
]



from __future__ import annotations

from dataclasses import dataclass

from rapidfuzz.distance import Levenshtein

from bewer.metrics.base import METRIC_REGISTRY, ExampleMetric, Metric, MetricParams, metric_value

__all__ = ["CER"]


class CER_(ExampleMetric):
    @metric_value
    def num_edits(self) -> int:
        """Get the number of edits between the hypothesis and reference text."""
        return Levenshtein.distance(
            self.example.hyp.joined(normalized=self.params.normalized),
            self.example.ref.joined(normalized=self.params.normalized),
        )

    @metric_value
    def ref_length(self) -> int:
        """Get the number of characters in the reference text."""
        return len(self.example.ref.joined(normalized=self.params.normalized))

    @metric_value(main=True)
    def value(self) -> float:
        """Get the example-level character error rate."""
        if self.ref_length == 0:
            return float(self.num_edits)
        return self.num_edits / self.ref_length


@METRIC_REGISTRY.register("cer")
class CER(Metric):
    short_name_base = "CER"
    long_name_base = "Character Error Rate"
    description = (
        "Character error rate (CER) is computed as the character-level edit distance between the reference "
        "and hypothesis texts, divided by the total number of characters in the reference texts."
    )
    example_cls = CER_

    @dataclass
    class param_schema(MetricParams):
        normalized: bool = True

    @metric_value
    def num_edits(self) -> int:
        """Get the number of edits between the hypothesis and reference texts."""
        return sum([self.get_example_metric(example).num_edits for example in self._src])

    @metric_value
    def ref_length(self) -> int:
        """Get the number of characters in the reference texts."""
        return sum([self.get_example_metric(example).ref_length for example in self._src])

    @metric_value(main=True)
    def value(self) -> float:
        """Get the character error rate."""
        if self.ref_length == 0:
            return float(self.num_edits)
        return self.num_edits / self.ref_length



from importlib.metadata import PackageNotFoundError, version

from bewer import core as core
from bewer import metrics as metrics
from bewer import preprocessing as preprocessing
from bewer import reporting as reporting
from bewer.core.dataset import Dataset
from bewer.core.keyword import KeywordNotFoundWarning

try:
    __version__ = version("bewer")
except PackageNotFoundError:
    __version__ = "0.0.0"

__all__ = [
    "Dataset",
    "KeywordNotFoundWarning",
    "core",
    "metrics",
    "preprocessing",
    "reporting",
    "__version__",
]



from bewer.core.dataset import Dataset, TextList, TextTokenList
from bewer.core.example import Example
from bewer.core.keyword import Keyword, KeywordNotFoundWarning
from bewer.core.text import Text, TextType, TokenList
from bewer.core.token import Token

__all__ = [
    "Dataset",
    "TextList",
    "TextTokenList",
    "Example",
    "KeywordNotFoundWarning",
    "Text",
    "TextType",
    "TokenList",
    "Token",
    "Keyword",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import MISSING, dataclass, fields
from functools import cached_property, update_wrapper
from typing import TYPE_CHECKING, Any, Optional, Union, get_type_hints

from typeguard import check_type

from bewer.flags import DEFAULT
from bewer.preprocessing.context import set_pipeline
from bewer.reporting.python.tables import print_metric_table

if TYPE_CHECKING:
    from bewer.core.dataset import Dataset
    from bewer.core.example import Example

__all__ = [
    "Metric",
    "ExampleMetric",
    "MetricParams",
    "metric_value",
    "METRIC_REGISTRY",
    "MetricRegistry",
    "list_registered_metrics",
    "MetricCollection",
    "ExampleMetricCollection",
]


class metric_value(cached_property):
    def __init__(self, func=None, *, main: bool = False):
        self.main = main
        if func is not None:
            super().__init__(func)
            update_wrapper(self, func)

    def __call__(self, func):
        super().__init__(func)
        update_wrapper(self, func)
        return self

    def __set_name__(self, owner: type, name: str):
        super().__set_name__(owner, name)

        # Create or get the _metric_values dict on the owner class
        metric_values = owner.__dict__.get("_metric_values")
        if metric_values is None:
            metric_values = {"other": [], "main": None}
            owner._metric_values = metric_values

        # Update the metric values
        if self.main:
            if metric_values["main"] is not None:
                raise ValueError(f"Multiple main metric values defined in {owner.__name__}.")
            metric_values["main"] = name
        else:
            metric_values["other"].append(name)

    def __get__(self, obj: Optional[Any], objtype: Optional[type] = None) -> Any:
        if obj is None:
            return self

        # TODO: Figure out if this is necessary given cached_property behavior
        name = self.attrname
        if name in obj.__dict__:
            return obj.__dict__[name]

        with set_pipeline(*obj.pipeline):
            value = self.func(obj)

        obj.__dict__[name] = value
        return value


def _get_metric_values(cls) -> dict[str, Union[str, list[str]]]:
    """Get the metric values defined in the class and its bases."""
    main_value = None
    other_values = []
    for base in reversed(cls.__mro__):
        _metric_values = base.__dict__.get("_metric_values")
        if _metric_values:
            main_value = _metric_values["main"] or main_value
            other_values.extend(_metric_values["other"])
    metric_values = {"main": main_value, "other": list(set(other_values))}
    return metric_values


def _get_metric_table_row_values(metric: "Metric") -> tuple[str, str, str]:
    metric_values = metric.metric_values()
    main_value = "-" if metric_values["main"] is None else metric_values["main"]
    other_values = "-" if len(metric_values["other"]) == 0 else ", ".join(metric_values["other"])
    return (main_value, other_values)


@dataclass
class MetricParams:
    """Base class for metric parameter dataclasses.

    Subclass this with @dataclass to define metric parameters:

        @dataclass
        class param_schema(MetricParams):
            threshold: float = 0.5

    The framework sets `_metric` after construction so that
    `self.metric` / `self.src` are available in `validate()`.
    """

    def __post_init__(self):
        hints = get_type_hints(type(self))
        for f in fields(self):
            value = getattr(self, f.name)
            expected_type = hints[f.name]
            try:
                check_type(value, expected_type)
            except Exception:
                type_name = getattr(expected_type, "__name__", str(expected_type))
                raise TypeError(f"Parameter '{f.name}' must be {type_name}, got {type(value).__name__}")

    @property
    def metric(self) -> "Metric":
        """Alias for the src property."""
        return self._metric

    @property
    def src(self) -> "Metric":
        """The parent Metric instance."""
        return self._metric

    def validate(self) -> None:
        """Override to validate params against the source metric/dataset.

        Called after set_source(), so self.metric.src is the Dataset.
        """
        pass


class Metric(ABC):
    example_cls: type["ExampleMetric"] | None = None
    param_schema: type["MetricParams"] | None = None
    short_name_base: str
    long_name_base: str

    def __init__(
        self,
        name: Optional[str] = None,
        src: Optional["Dataset"] = None,
        *,
        standardizer: str = DEFAULT,
        tokenizer: str = DEFAULT,
        normalizer: str = DEFAULT,
        **params,
    ):
        """Initialize the Metric object.

        Args:
            name: Metric name. Defaults to the lowercase class name.
            src: Dataset object for computing the metric. Can be set later via set_source().
            standardizer: Standardizer pipeline name.
            tokenizer: Tokenizer pipeline name.
            normalizer: Normalizer pipeline name.
            **params: Optional parameters for metric configuration.
        """
        self.name = name or type(self).__name__.lower()
        self._examples = {}

        self._standardizer = standardizer
        self._tokenizer = tokenizer
        self._normalizer = normalizer
        self._pipeline = (standardizer, tokenizer, normalizer)

        # Construct params dataclass or reject unexpected params
        if self.param_schema is not None:
            try:
                self.params = self.param_schema(**params)
            except TypeError as e:
                dc_fields = {f.name for f in fields(self.param_schema)}
                unknown = set(params.keys()) - dc_fields
                if unknown:
                    raise ValueError(
                        f"Unknown parameter(s) {sorted(unknown)} for {self.short_name_base}. "
                        f"Valid parameters: {sorted(dc_fields)}"
                    ) from e
                missing = {
                    f.name for f in fields(self.param_schema) if f.default is MISSING and f.default_factory is MISSING
                } - set(params.keys())
                if missing:
                    param_hints = ", ".join(f"{p}=..." for p in sorted(missing))
                    raise ValueError(
                        f"Missing required parameters for {self.short_name_base}: {sorted(missing)}. "
                        f"Pass them as keyword arguments: metric({param_hints})"
                    ) from e
                raise
            self.params._metric = self
        else:
            if params:
                raise ValueError(f"Metric {self.short_name_base} does not accept parameters")
            self.params = None

        self._src = None
        if src is not None:
            self.set_source(src)

    @property
    def short_name(self) -> str:
        """Get the short name, including parameters if present."""
        if self.params is None:
            return self.short_name_base
        param_strs = [f"{f.name}={getattr(self.params, f.name)}" for f in fields(self.params)]
        return f"{self.short_name_base} ({', '.join(param_strs)})"

    @property
    def long_name(self) -> str:
        """Get the long name, including parameters if present."""
        if self.params is None:
            return self.long_name_base
        param_strs = [f"{f.name}={getattr(self.params, f.name)}" for f in fields(self.params)]
        return f"{self.long_name_base} ({', '.join(param_strs)})"

    @property
    @abstractmethod
    def description(self) -> str:
        """Get a description of the metric."""
        pass

    @property
    def src(self) -> Optional["Dataset"]:
        """Get the parent Dataset object."""
        return self._src

    @property
    def dataset(self) -> Optional["Dataset"]:
        """Alias for src property."""
        return self._src

    @property
    def pipeline(self) -> tuple[str, str, str]:
        """Get the preprocessing pipeline for the metric. Cached property to ensure immutability."""
        return self._pipeline

    @property
    def standardizer(self) -> str:
        """Get the standardizer for the metric."""
        return self._standardizer

    @property
    def tokenizer(self) -> str:
        """Get the tokenizer for the metric."""
        return self._tokenizer

    @property
    def normalizer(self) -> str:
        """Get the normalizer for the metric."""
        return self._normalizer

    @classmethod
    def metric_values(cls) -> dict[str, Union[str, list[str]]]:
        """Get the metric values defined in the class and its bases."""
        return _get_metric_values(cls)

    @classmethod
    def _get_row_values(cls) -> tuple[str, str, str] | None:
        """Get the table row values for the main and example metric."""
        metric_row_values = _get_metric_table_row_values(cls)
        if cls.example_cls is not None:
            example_metric_row_values = _get_metric_table_row_values(cls.example_cls)
        else:
            example_metric_row_values = None
        return (metric_row_values, example_metric_row_values)

    def set_source(self, src: "Dataset") -> None:
        """Set the parent Dataset object and validate parameters if needed.

        Args:
            src: The parent Dataset object.

        Raises:
            ValueError: If source is already set or if parameters are invalid for this dataset.
        """
        if self._src is not None:
            raise ValueError("Source already set for Metric")
        self._src = src

        if self.params is not None:
            self.params.validate()

    def get_example_metric(self, example: "Example") -> "ExampleMetric":
        """Get the ExampleMetric object for a given example index."""
        if example._index in self._examples:
            return self._examples[example._index]
        if self.example_cls is None:
            return None
        example_metric = self.example_cls(parent_metric=self)
        example_metric.set_source(example)
        self._examples[example._index] = example_metric
        return example_metric


class ExampleMetric(ABC):
    def __init__(
        self,
        parent_metric: "Metric",
        src: Optional["Example"] = None,
    ):
        """Initialize the ExampleMetric object.

        Args:
            parent_metric: The parent Metric object.
            src: Parent Example object. Can be set later via set_source().
        """
        self.parent_metric = parent_metric

        self._src = None
        if src is not None:
            self.set_source(src)

    @property
    def params(self) -> Optional["MetricParams"]:
        """Access parent metric's parameters."""
        return self.parent_metric.params

    @property
    def src(self) -> Optional["Example"]:
        """Get the parent Example object."""
        return self._src

    @property
    def example(self) -> Optional["Example"]:
        """Alias for src property."""
        return self._src

    @property
    def pipeline(self) -> tuple[str, str, str]:
        """Get the preprocessing pipeline for the metric."""
        return self.parent_metric.pipeline

    @property
    def standardizer(self) -> str:
        """Get the standardizer for the metric."""
        return self.parent_metric.standardizer

    @property
    def tokenizer(self) -> str:
        """Get the tokenizer for the metric."""
        return self.parent_metric.tokenizer

    @property
    def normalizer(self) -> str:
        """Get the normalizer for the metric."""
        return self.parent_metric.normalizer

    def set_source(self, src: "Example") -> None:
        """Set the parent Example object.

        Args:
            src: The parent Example object.

        Raises:
            ValueError: If source is already set.
        """
        if self._src is not None:
            raise ValueError("Source already set for ExampleMetric")
        self._src = src

    @classmethod
    def metric_values(cls) -> dict[str, Union[str, list[str]]]:
        """Get the metric values defined in the class and its bases."""
        return _get_metric_values(cls)


class MetricCollection(object):
    """Collection of metrics for a dataset or an example.

    Attributes:
        src (Union[Dataset, Example]): The source object (dataset or example) to compute metrics for.
        metrics (list[Metric]): A list of valid Metric objects depending on the source object.
    """

    def __init__(self, src: "Dataset"):
        """Initialize the MetricCollection object.

        Args:
            src (Union[Dataset, Example]): The source object (dataset or example) to compute metrics for.
        """
        self._src = src
        self._metric_cache = {}  # (name, kwargs_key) -> Metric instance

    @staticmethod
    def _make_cache_key(**kwargs) -> tuple:
        """Create a hashable cache key from kwargs.

        Args:
            **kwargs: Keyword arguments to hash. All values must be hashable.

        Returns:
            Tuple of sorted (key, value) pairs.
        """
        if not kwargs:
            return ()
        return tuple(sorted(kwargs.items()))

    def list_metrics(self, show_private: bool = False) -> None:
        """Print all registered example metric and their values."""
        metric_rows = []
        for metric_name, metric_cls in METRIC_REGISTRY.metric_classes.items():
            if not show_private and metric_name.startswith("_"):
                continue
            metric_rows.append((metric_name, metric_cls._get_row_values()))
        print_metric_table(metric_rows)

    def get(self, name: str):
        """Get a metric factory function by name.

        Returns a callable that creates/caches metric instances with specified parameters.

        Args:
            name: The registered metric name.

        Returns:
            A factory function that accepts **kwargs and returns a Metric instance.

        Example:
            >>> wer_metric = dataset.metrics.get("wer")()
            >>> wer_value = wer_metric.value
        """
        if name not in METRIC_REGISTRY.metric_metadata:
            raise AttributeError(f"Metric '{name}' not found.")

        # Return factory function
        def metric_factory(**kwargs):
            # Resolve kwargs against all defaults for a canonical cache key
            resolved = METRIC_REGISTRY.resolve_params(name, **kwargs)
            try:
                cache_key = (name, self._make_cache_key(**resolved))
                if cache_key in self._metric_cache:
                    return self._metric_cache[cache_key]
            except TypeError as e:
                # Find non-hashable parameters for better error message
                non_hashable = []
                for key, value in kwargs.items():
                    try:
                        hash(value)
                    except TypeError:
                        non_hashable.append(f"{key} ({type(value).__name__})")
                raise TypeError(
                    f"All metric parameters must be hashable. Non-hashable parameters: {', '.join(non_hashable)}. "
                    f"Use hashable alternatives: tuple instead of list, frozenset instead of set, etc."
                ) from e

            # Create new metric instance
            metric_instance = METRIC_REGISTRY.create_metric(name, **kwargs)
            metric_instance.set_source(self._src)

            # Cache and return
            self._metric_cache[cache_key] = metric_instance
            return metric_instance

        return metric_factory

    def __getattr__(self, name: str):
        # NOTE: This method is only called if the attribute is not found the usual ways.
        return self.get(name)

    def __repr__(self):
        # TODO: Improve representation to list available metrics and their values.
        return "MetricCollection()"


class ExampleMetricCollection(object):
    """Collection of metrics for an example."""

    def __init__(self, src: "Example"):
        """Initialize the ExampleMetricCollection object."""
        self._src_example = src
        self._src_collection = src.src.metrics if src.src is not None else None
        self._cache = {}

    def get(self, name: str):
        """Get a metric factory function by name.

        Returns a callable that creates/caches example metric instances.

        Args:
            name: The registered metric name.

        Returns:
            A factory function that accepts **kwargs and returns an ExampleMetric instance.
        """
        if name not in METRIC_REGISTRY.metric_metadata:
            raise AttributeError(f"Metric '{name}' not found.")

        # Return factory function
        def example_metric_factory(**kwargs):
            # Resolve kwargs against all defaults for a canonical cache key
            resolved = METRIC_REGISTRY.resolve_params(name, **kwargs)
            try:
                cache_key = (name, MetricCollection._make_cache_key(**resolved))
                if cache_key in self._cache:
                    return self._cache[cache_key]
            except TypeError as e:
                non_hashable = []
                for key, value in kwargs.items():
                    try:
                        hash(value)
                    except TypeError:
                        non_hashable.append(f"{key} ({type(value).__name__})")
                raise TypeError(
                    f"All metric parameters must be hashable. Non-hashable parameters: {', '.join(non_hashable)}. "
                    f"Use hashable alternatives: tuple instead of list, frozenset instead of set, etc."
                ) from e

            # Get the parent metric (with same params)
            parent_metric_factory = self._src_collection.__getattr__(name)
            parent_metric = parent_metric_factory(**kwargs)

            # Get example metric from parent
            example_metric_instance = parent_metric.get_example_metric(self._src_example)

            # Cache and return
            self._cache[cache_key] = example_metric_instance
            return example_metric_instance

        return example_metric_factory

    def __getattr__(self, name: str):
        # NOTE: This method is only called if the attribute is not found the usual ways.
        return self.get(name)

    def __repr__(self):
        # TODO: Improve representation to list available metrics and their values.
        return "ExampleMetricCollection()"


class MetricRegistry:
    def __init__(self) -> None:
        self.metric_metadata = {}  # name -> metadata dict
        self.metric_classes = {}  # name -> class (kept for compatibility)

    @property
    def metric_factories(self):
        """Backwards compatibility - returns dict of names."""
        return self.metric_metadata

    def register_metric(
        self,
        metric_cls: type["Metric"],
        name: str,
        allow_override: bool = False,
        standardizer: str = DEFAULT,
        tokenizer: str = DEFAULT,
        normalizer: str = DEFAULT,
        **kwargs,
    ):
        """
        Register a metric.

        Args:
            metric_cls (type): The metric class to register.
            name (str): The registered name for the metric.
            allow_override (bool): Whether to allow overriding an existing metric with the same name.
            standardizer (str): Default standardizer for the metric.
            tokenizer (str): Default tokenizer for the metric.
            normalizer (str): Default normalizer for the metric.
            **kwargs: Default parameters to pass to the metric class upon instantiation.
        """

        # Validate metric class.
        if not issubclass(metric_cls, Metric):
            raise TypeError(f"Metric class {metric_cls.__name__} must inherit from Metric.")

        # Validate metric name.
        if not isinstance(name, str):
            raise TypeError("Metric name must be a string.")

        # Register metric based on its type.
        if name in self.metric_metadata and not allow_override:
            raise ValueError(f"Metric '{name}' already registered.")

        # Extract parameter schema from dataclass fields
        if metric_cls.param_schema is not None:
            if not hasattr(metric_cls.param_schema, "__dataclass_fields__"):
                raise TypeError(f"param_schema on {metric_cls.__name__} must be a @dataclass inheriting MetricParams.")
            hints = get_type_hints(metric_cls.param_schema)
            param_schema = {}
            for f in fields(metric_cls.param_schema):
                field_type = hints[f.name]
                if f.default is not MISSING:
                    param_schema[f.name] = (field_type, f.default)
                elif f.default_factory is not MISSING:
                    param_schema[f.name] = (field_type, f.default_factory)
                else:
                    param_schema[f.name] = field_type
        else:
            param_schema = {}

        # Determine required parameters (those without defaults)
        required_params = {
            param_name
            for param_name, param_spec in param_schema.items()
            if not isinstance(param_spec, tuple)  # No tuple = no default
        }

        # Store metadata for factory creation
        metadata = {
            "metric_cls": metric_cls,
            "pipeline_defaults": {
                "standardizer": standardizer,
                "tokenizer": tokenizer,
                "normalizer": normalizer,
            },
            "param_defaults": kwargs,  # Params passed at registration
            "param_schema": param_schema,
            "required_params": required_params,
        }

        self.metric_metadata[name] = metadata
        self.metric_classes[name] = metric_cls

    def resolve_params(self, name: str, **kwargs) -> dict:
        """Resolve kwargs against all defaults to get canonical form.

        Merges pipeline defaults, registry param defaults, and class-level param_schema
        defaults with provided kwargs to produce a fully-resolved parameter dict.

        Args:
            name: The registered metric name.
            **kwargs: Raw kwargs (pipeline overrides + metric params).

        Returns:
            Dict with all resolved key-value pairs (pipeline + metric params).
        """
        metadata = self.metric_metadata[name]

        pipeline_args = {}
        metric_params = {}
        for key, value in kwargs.items():
            if key in ("standardizer", "tokenizer", "normalizer"):
                pipeline_args[key] = value
            else:
                metric_params[key] = value

        # Merge with registry-level defaults
        final_pipeline = {**metadata["pipeline_defaults"], **pipeline_args}
        final_params = {**metadata["param_defaults"], **metric_params}

        # Apply class-level param_schema defaults for any still-missing params
        for param_name, param_spec in metadata["param_schema"].items():
            if isinstance(param_spec, tuple) and param_name not in final_params:
                default = param_spec[1]
                final_params[param_name] = default() if callable(default) else default

        return {**final_pipeline, **final_params}

    def create_metric(self, name: str, **kwargs) -> "Metric":
        """Create a metric instance with merged defaults and overrides.

        Args:
            name: The registered metric name.
            **kwargs: Parameters and pipeline overrides. Can include:
                - standardizer, tokenizer, normalizer (pipeline overrides)
                - Any metric-specific parameters

        Returns:
            Configured Metric instance.
        """
        if name not in self.metric_metadata:
            raise ValueError(f"Metric '{name}' not registered.")

        resolved = self.resolve_params(name, **kwargs)

        # Separate pipeline args from metric params
        pipeline_keys = ("standardizer", "tokenizer", "normalizer")
        pipeline_args = {k: resolved[k] for k in pipeline_keys}
        metric_params = {k: v for k, v in resolved.items() if k not in pipeline_keys}

        return self.metric_metadata[name]["metric_cls"](name=name, **pipeline_args, **metric_params)

    def register(
        self,
        name: str | None = None,
        allow_override: bool = False,
        standardizer: str = DEFAULT,
        tokenizer: str = DEFAULT,
        normalizer: str = DEFAULT,
        **kwargs,
    ):
        """
        Decorator version of 'register_metric'.

        Usage:
            @registry.register("foo")
            class Foo(...): ...

            @registry.register(name="foo", bar=3)
            class Foo(...): ...
        """

        def decorator(metric_cls):
            # Delegate to the existing register() method
            self.register_metric(
                metric_cls=metric_cls,
                name=name,
                allow_override=allow_override,
                standardizer=standardizer,
                tokenizer=tokenizer,
                normalizer=normalizer,
                **kwargs,
            )
            return metric_cls

        return decorator


METRIC_REGISTRY = MetricRegistry()


def list_registered_metrics(show_private: bool = False) -> list[str]:
    """List all registered metric names.

    Args:
        show_private (bool): Whether to include private metrics (those starting with an underscore).

    Returns:
        list[str]: List of registered metric names.
    """
    if show_private:
        return list(METRIC_REGISTRY.metric_factories.keys())
    else:
        return [name for name in METRIC_REGISTRY.metric_factories.keys() if not name.startswith("_")]



__all__ = ["HTMLAlignmentColors", "HTMLDefaultAlignmentColors", "HTMLBaseColors"]


class HTMLAlignmentColors:
    """Base class for HTML color schemes used in alignment display."""

    PAD: str
    DEL: str
    INS: str
    SUB: str
    MATCH: str
    KEYWORD: str


class HTMLDefaultAlignmentColors(HTMLAlignmentColors):
    """Default color scheme for HTML alignment display."""

    PAD = "#ededed"
    DEL = "#dc3545"
    INS = "#17a2b8"
    SUB = "#c46f01"
    MATCH = "#212529"
    KEYWORD = "#00a2ff"


class HTMLBaseColors:
    """Color scheme constants for HTML reports."""

    TABLE_HEADER_BG = "#595959"
    TEXT_COLOR = "#000000"
    BG_COLOR = "#faf9f5"
    DIV_TEXT_COLOR = "#acacac"
    DIV_BG_COLOR = "#ffffff"



"""Module for styling alignment display as HTML.

Alignments can be rendered as HTML with color coding for different operation types.
The generated HTML can be saved to a file and viewed in a browser.
"""

from html import escape
from typing import TYPE_CHECKING

from bewer.alignment.op_type import OpType
from bewer.reporting.html.color_schemes import (
    HTMLAlignmentColors,
    HTMLDefaultAlignmentColors,
)

if TYPE_CHECKING:
    from bewer.alignment.alignment import Alignment
    from bewer.alignment.op import Op

__all__ = ["generate_alignment_html_lines"]


def get_html_padding(length: int, color_scheme: type[HTMLAlignmentColors] = HTMLDefaultAlignmentColors) -> str:
    """Get an HTML span representing padding spaces.

    Args:
        length: The number of spaces for padding.
        color_scheme: The color scheme to use.

    Returns:
        An HTML span element with the padding.
    """
    spaces = "&nbsp;" * length
    return f'<span style="background-color: {color_scheme.PAD};">{spaces}</span>'


def format_match_op_html(
    op: "Op", color_scheme: type[HTMLAlignmentColors] = HTMLDefaultAlignmentColors
) -> tuple[str, str, int]:
    """Format a match operation for HTML display."""
    len_ref = len(op.ref)
    len_hyp = len(op.hyp)
    length = max(len_ref, len_hyp)

    ref_str = f'<span style="color: {color_scheme.MATCH};">{escape(op.ref)}</span>'
    hyp_str = f'<span style="color: {color_scheme.MATCH};">{escape(op.hyp)}</span>'

    if len_ref < length:
        ref_str += get_html_padding(length - len_ref, color_scheme=color_scheme)
    if len_hyp < length:
        hyp_str += get_html_padding(length - len_hyp, color_scheme=color_scheme)

    return ref_str, hyp_str, length


def format_substitute_op_html(
    op: "Op", color_scheme: type[HTMLAlignmentColors] = HTMLDefaultAlignmentColors
) -> tuple[str, str, int]:
    """Format a substitute operation for HTML display."""
    len_ref = len(op.ref)
    len_hyp = len(op.hyp)
    length = max(len_ref, len_hyp)

    ref_str = f'<span style="color: {color_scheme.SUB};">{escape(op.ref)}</span>'
    hyp_str = f'<span style="color: {color_scheme.SUB};">{escape(op.hyp)}</span>'

    if len_ref < length:
        ref_str += get_html_padding(length - len_ref, color_scheme=color_scheme)
    if len_hyp < length:
        hyp_str += get_html_padding(length - len_hyp, color_scheme=color_scheme)

    return ref_str, hyp_str, length


def format_insert_op_html(
    op: "Op", color_scheme: type[HTMLAlignmentColors] = HTMLDefaultAlignmentColors
) -> tuple[str, str, int]:
    """Format an insert operation for HTML display."""
    len_hyp = len(op.hyp)
    hyp_str = f'<span style="color: {color_scheme.INS};">{escape(op.hyp)}</span>'
    ref_str = get_html_padding(len_hyp, color_scheme=color_scheme)
    return ref_str, hyp_str, len_hyp


def format_delete_op_html(
    op: "Op", color_scheme: type[HTMLAlignmentColors] = HTMLDefaultAlignmentColors
) -> tuple[str, str, int]:
    """Format a delete operation for HTML display."""
    len_ref = len(op.ref)
    ref_str = f'<span style="color: {color_scheme.DEL};">{escape(op.ref)}</span>'
    hyp_str = get_html_padding(len_ref, color_scheme=color_scheme)
    return ref_str, hyp_str, len_ref


def format_alignment_op_html(
    op: "Op", color_scheme: type[HTMLAlignmentColors] = HTMLDefaultAlignmentColors
) -> tuple[str, str, int]:
    """Format an alignment operation for HTML display.

    Args:
        op: The alignment operation.
        color_scheme: The color scheme to use.

    Returns:
        A tuple containing the formatted ref and hyp HTML strings and the unformatted length.
    """
    if op.type == OpType.MATCH:
        return format_match_op_html(op, color_scheme=color_scheme)
    if op.type == OpType.SUBSTITUTE:
        return format_substitute_op_html(op, color_scheme=color_scheme)
    if op.type == OpType.INSERT:
        return format_insert_op_html(op, color_scheme=color_scheme)
    if op.type == OpType.DELETE:
        return format_delete_op_html(op, color_scheme=color_scheme)
    raise ValueError(f"Unknown operation type: {op.type}")


def format_keyword(text: str, start: bool = False, end: bool = False) -> str:
    """Format a keyword with HTML tags for highlighting.

    Args:
        text: The keyword text to format.
        start: Whether this is the start of a keyword span.
        end: Whether this is the end of a keyword span.

    Returns:
        The formatted keyword string with HTML span tags.
    """
    kw_class = "kw"
    if start:
        kw_class += " kw-start"
    if end:
        kw_class += " kw-end"
    return f'<span class="{kw_class}">{text}</span>'


def _get_keyword_indicators(alignment: "Alignment") -> tuple[set[int], set[int], set[int]]:
    """Compute keyword span indicators for the given alignment.

    Args:
        alignment: The alignment whose reference side is inspected for keyword spans.

    Returns:
        A tuple of three sets of operation indices:

        - start_indices: Indices of alignment operations that correspond to the first
          token of a keyword span in the reference text.
        - stop_indices: Indices of alignment operations that correspond to the last
          token of a keyword span in the reference text.
        - open_indices: Indices of alignment operations that fall inside any keyword
          span (including the start index but excluding the stop index), i.e. where a
          keyword span is considered "open"/ongoing.
    """
    example = alignment.src
    if example is None:
        return set(), set(), set()
    vocabs = example.vocabs
    if not vocabs:
        return set(), set(), set()

    start_indices, stop_indices, open_indices = set(), set(), set()
    for vocab in vocabs:
        matches = example.get_keyword_matches(vocab=vocab)
        for match in matches:
            start_op_idx = alignment.ref_index_mapping.get(match.start)
            end_op_idx = alignment.ref_index_mapping.get(match.stop - 1)
            if start_op_idx is None or end_op_idx is None:
                continue
            start_indices.add(start_op_idx)
            stop_indices.add(end_op_idx)
            for idx in range(start_op_idx, end_op_idx):
                open_indices.add(idx)

    return start_indices, stop_indices, open_indices


def generate_alignment_html_lines(
    alignment: "Alignment",
    max_line_length: int = 100,
    color_scheme: type[HTMLAlignmentColors] = HTMLDefaultAlignmentColors,
) -> list[tuple[str, str]]:
    """Render the alignment as an HTML table.

    This function generates only the inner content of the alignment container,
    without the full HTML document wrapper. Use this for embedding alignments
    in templates or combining multiple alignments.

    Args:
        alignment: The alignment to render.
        max_line_length: The maximum character length per line for wrapping.
        color_scheme: The color scheme to use for display.

    Returns:
        A list of tuples, each containing the reference and hypothesis HTML strings for each line.
    """
    ref_line, hyp_line = "", ""
    current_length = 0

    start_indices, stop_indices, open_indices = _get_keyword_indicators(alignment)

    lines = []
    for op_idx, op in enumerate(alignment):
        ref_str, hyp_str, op_length = format_alignment_op_html(op, color_scheme=color_scheme)

        is_kw_start = op_idx in start_indices
        is_kw_end = op_idx in stop_indices
        is_kw_open = op_idx in open_indices
        is_kw = is_kw_start or is_kw_end or is_kw_open
        if is_kw:
            ref_str = format_keyword(ref_str, start=is_kw_start, end=is_kw_end)

        if current_length + op_length > max_line_length and current_length > 0:
            lines.append((ref_line, hyp_line))
            ref_line, hyp_line = "", ""
            current_length = 0

        ref_line += ref_str + (format_keyword("&nbsp;") if is_kw_open else "&nbsp;")
        hyp_line += hyp_str + (get_html_padding(1, color_scheme=color_scheme) if op.hyp_right_partial else "&nbsp;")
        current_length += op_length + 1

    lines.append((ref_line, hyp_line))
    return lines



"""Module for generating HTML reports from datasets."""

from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING

from jinja2 import Environment, PackageLoader

from bewer.reporting.html.color_schemes import HTMLAlignmentColors, HTMLBaseColors, HTMLDefaultAlignmentColors
from bewer.reporting.html.labels import HTMLAlignmentLabels

if TYPE_CHECKING:
    from bewer.core.dataset import Dataset

__all__ = [
    "generate_report",
    "render_report_html",
    "ReportMetric",
    "ReportAlignment",
    "ReportSummaryItem",
    "DEFAULT_REPORT_METRICS",
    "DEFAULT_REPORT_SUMMARY_ITEMS",
    "DEFAULT_REPORT_ALIGNMENT",
]


class ReportMetric:
    """Specification for a metric to include in the report."""

    def __init__(self, name: str, label: str | None = None, format: str = ".2%", **metric_kwargs):
        self.name = name  # metric registry name (e.g. "wer")
        self.label = label  # display label override (default: metric.long_name)
        self.format = format  # format spec for the value
        self.metric_kwargs = metric_kwargs  # optional kwargs to pass when resolving the metric from the dataset


class ReportAlignment:
    """Specification for the alignment to include in the report."""

    def __init__(self, name: str, **kwargs):
        self.name = name  # metric registry name (e.g. "levenshtein")
        self.metric_kwargs = kwargs  # kwargs passed to the metric factory


class ReportSummaryItem:
    """Specification for a summary item to include in the report."""

    def __init__(self, name: str, label: str | None = None, format: str = ",.0f"):
        self.name = name  # summary attribute name (e.g. "num_examples")
        self.label = label  # display label override
        self.format = format  # format spec


DEFAULT_REPORT_ALIGNMENT = ReportAlignment("levenshtein")

DEFAULT_REPORT_METRICS = [
    ReportMetric("wer"),
    ReportMetric("cer"),
    ReportMetric("legacy_medical_word_accuracy", label="Medical Term Recall"),
]

DEFAULT_REPORT_SUMMARY_ITEMS = [
    ReportSummaryItem("num_examples", label="Number of examples"),
    ReportSummaryItem("num_ref_words", label="Number of reference words"),
    ReportSummaryItem("num_ref_chars", label="Number of reference characters"),
    ReportSummaryItem("num_hyp_words", label="Number of hypothesis words"),
    ReportSummaryItem("num_hyp_chars", label="Number of hypothesis characters"),
]


def indent_tabs(text: str, width: int = 1) -> str:
    """Indent each line of the given text with a specified number of tabs."""
    padding = "\t" * width
    return "\n".join(padding + line for line in text.split("\n"))


def render_report_html(
    dataset: "Dataset",
    template: str = "report_basic",
    title: str | None = None,
    base_color_scheme: type[HTMLBaseColors] = HTMLBaseColors,
    alignment_color_scheme: type[HTMLAlignmentColors] = HTMLDefaultAlignmentColors,
    alignment_labels: type[HTMLAlignmentLabels] = HTMLAlignmentLabels,
    report_metrics: list[ReportMetric] | None = None,
    report_summary: list[ReportSummaryItem] | None = None,
    report_alignment: ReportAlignment | None = None,
    metadata: dict[str, str] | None = None,
) -> str:
    """Render an HTML report with alignment visualizations for all examples in a dataset.

    Args:
        dataset: The dataset to generate the report for.
        template: The template name to use (e.g., "report_basic"). Templates are looked up in the bewer.templates
            package.
        title: An optional title for the report.
        base_color_scheme: The base color scheme to use for the report.
        alignment_color_scheme: The color scheme to use for alignment display.
        alignment_labels: The labels and tooltips to use for alignment display.
        report_metrics: List of ReportMetric specs controlling which metrics appear. Defaults to
            DEFAULT_REPORT_METRICS.
        report_summary: List of ReportSummaryItem specs controlling the summary section. Defaults to
            DEFAULT_REPORT_SUMMARY_ITEMS.
        report_alignment: ReportAlignment spec controlling which alignment to display. Defaults to
            DEFAULT_REPORT_ALIGNMENT.
        metadata: Optional dict of key-value pairs to display in the report metadata line.

    Returns:
        The rendered HTML report string.
    """
    if report_metrics is None:
        report_metrics = DEFAULT_REPORT_METRICS
    if report_summary is None:
        report_summary = DEFAULT_REPORT_SUMMARY_ITEMS
    if report_alignment is None:
        report_alignment = DEFAULT_REPORT_ALIGNMENT
    if metadata is None:
        metadata = {}

    # Resolve metrics against the dataset
    resolved_metrics = []
    for spec in report_metrics:
        metric = dataset.metrics.get(spec.name)(**spec.metric_kwargs)
        label = spec.label if spec.label is not None else metric.long_name
        resolved_metrics.append({"name": label, "value": f"{metric.value:{spec.format}}"})

    # Resolve summary items against the dataset summary
    resolved_summary = []
    for spec in report_summary:
        value = getattr(dataset.metrics.summary(), spec.name)
        label = spec.label if spec.label is not None else spec.name
        resolved_summary.append({"name": label, "value": f"{value:{spec.format}}"})

    # Resolve alignments for each example
    resolved_alignments = []
    for example in dataset:
        alignment = example.metrics.get(report_alignment.name)(**report_alignment.metric_kwargs).alignment
        resolved_alignments.append(alignment)

    # Load and render the Jinja template
    env = Environment(loader=PackageLoader("bewer", "templates"), autoescape=True)
    env.filters["indent_tabs"] = indent_tabs
    jinja_template = env.get_template(f"{template}.html.j2")

    html = jinja_template.render(
        dataset=dataset,
        title=title,
        creation_date=datetime.now().strftime("%B %d, %Y"),
        base_color_scheme=base_color_scheme,
        metrics=resolved_metrics,
        summary=resolved_summary,
        alignments=resolved_alignments,
        alignment_color_scheme=alignment_color_scheme,
        alignment_labels=alignment_labels,
        metadata=metadata,
    )

    return html


def generate_report(
    dataset: "Dataset",
    path: str | Path | None = None,
    allow_overwrite: bool = False,
    template: str = "report_basic",
    title: str | None = None,
    base_color_scheme: type[HTMLBaseColors] = HTMLBaseColors,
    alignment_color_scheme: type[HTMLAlignmentColors] = HTMLDefaultAlignmentColors,
    alignment_labels: type[HTMLAlignmentLabels] = HTMLAlignmentLabels,
    report_metrics: list[ReportMetric] | None = None,
    report_summary: list[ReportSummaryItem] | None = None,
    report_alignment: ReportAlignment | None = None,
    metadata: dict[str, str] | None = None,
) -> str:
    """Generate an HTML report with alignment visualizations for all examples.

    Args:
        dataset: The dataset to generate the report for.
        path: If provided, write the HTML to this file.
        allow_overwrite: If True, overwrite the file if it exists.
        template: The template name to use (e.g., "report_basic"). Templates are looked up
            in the bewer.templates package.
        title: An optional title for the report.
        base_color_scheme: The base color scheme to use for the report.
        alignment_color_scheme: The color scheme to use for alignment display.
        alignment_labels: The labels and tooltips to use for alignment display.
        report_metrics: List of ReportMetric specs controlling which metrics appear. Defaults to
            DEFAULT_REPORT_METRICS.
        report_summary: List of ReportSummaryItem specs controlling the summary section. Defaults to
            DEFAULT_REPORT_SUMMARY_ITEMS.
        report_alignment: ReportAlignment spec controlling which alignment to display. Defaults to
            DEFAULT_REPORT_ALIGNMENT.
        metadata: Optional dict of key-value pairs to display in the report metadata line.

    Returns:
        The rendered HTML report string.
    """
    html = render_report_html(
        dataset,
        template=template,
        title=title,
        base_color_scheme=base_color_scheme,
        alignment_color_scheme=alignment_color_scheme,
        alignment_labels=alignment_labels,
        report_metrics=report_metrics,
        report_summary=report_summary,
        report_alignment=report_alignment,
        metadata=metadata,
    )

    if path is not None:
        path = Path(path)
        if path.is_dir():
            raise ValueError("Provided path is a directory, expected a file path.")
        if path.exists() and not allow_overwrite:
            raise FileExistsError(f"File {path} already exists.")
        if not path.parent.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "w", encoding="utf-8") as f:
            f.write(html)

    return html



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from error_align import error_align
from error_align.utils import OpType as EAOpType
from error_align.utils import basic_normalizer, basic_tokenizer

from bewer.alignment import Alignment, Op, OpType
from bewer.metrics.base import METRIC_REGISTRY, ExampleMetric, Metric, MetricParams, metric_value
from bewer.preprocessing.context import get_normalizer, get_tokenizer

__all__ = ["ErrorAlign"]


class ErrorAlign_(ExampleMetric):
    """Compute ErrorAlign edit operations between hypothesis and reference text."""

    OPS_MAP = {
        EAOpType.MATCH: OpType.MATCH,
        EAOpType.SUBSTITUTE: OpType.SUBSTITUTE,
        EAOpType.INSERT: OpType.INSERT,
        EAOpType.DELETE: OpType.DELETE,
    }

    @metric_value
    def num_substitutions(self) -> int:
        """Get the number of substitutions."""
        return self.alignment.num_substitutions

    @metric_value
    def num_insertions(self) -> int:
        """Get the number of insertions."""
        return self.alignment.num_insertions

    @metric_value
    def num_deletions(self) -> int:
        """Get the number of deletions."""
        return self.alignment.num_deletions

    @metric_value
    def num_edits(self) -> int:
        """Get the number of edits."""
        return self.alignment.num_edits

    @metric_value
    def num_matches(self) -> int:
        """Get the number of matches."""
        return self.alignment.num_matches

    @metric_value(main=True)
    def alignment(self) -> Alignment:
        """Get the error alignment operations between the hypothesis and reference text."""
        return self._get_ops()

    @staticmethod
    def _normalize_conditionally(text: Optional[str], normalizer: Optional[callable]) -> Optional[str]:
        """Normalize text if normalizer is provided, otherwise return original text."""
        if text is None:
            return None
        return normalizer(text) if normalizer else text

    @staticmethod
    def _no_normalizer(text: str) -> str:
        """Return text unchanged (used when no normalizer is provided)."""
        return text

    def _get_ops(self) -> list[Op]:
        """
        Compute and convert ErrorAlign edit operations to BeWER operations.

        Returns:
            list[Op]: List of BeWER operations.
        """
        tokenizer = get_tokenizer(self.parent_metric.dataset)
        tokenizer = tokenizer or basic_tokenizer
        normalizer = get_normalizer(self.parent_metric.dataset) if self.params.normalized else None
        ea_ops = []
        ref_idx = 0
        for ea_op in error_align(
            self.example.ref.standardized,
            self.example.hyp.standardized,
            tokenizer=tokenizer,
            normalizer=basic_normalizer if self.params.normalized else self._no_normalizer,
        ):
            ref_empty = ea_op.ref is None
            op = Op(
                type=self.OPS_MAP[ea_op.op_type],
                ref=self._normalize_conditionally(ea_op.ref, normalizer),
                hyp=self._normalize_conditionally(ea_op.hyp, normalizer),
                ref_token_idx=None if ref_empty else ref_idx,
                hyp_token_idx=None,
                ref_span=ea_op.ref_slice,
                hyp_span=ea_op.hyp_slice,
                hyp_left_partial=ea_op.left_compound,
                hyp_right_partial=ea_op.right_compound,
            )
            if not ref_empty:
                ref_idx += 1
            ea_ops.append(op)

        alignment = Alignment(ea_ops, src=self.example)
        return alignment


@METRIC_REGISTRY.register("error_align")
class ErrorAlign(Metric):
    short_name_base = "EA"
    long_name_base = "Error Alignment"
    description = "Error alignment between hypothesis and reference texts."
    example_cls = ErrorAlign_

    @dataclass
    class param_schema(MetricParams):
        normalized: bool = True



__all__ = [
    "generate_report",
    "render_report_html",
    "ReportMetric",
    "ReportAlignment",
    "ReportSummaryItem",
    "DEFAULT_REPORT_METRICS",
    "DEFAULT_REPORT_SUMMARY_ITEMS",
    "DEFAULT_REPORT_ALIGNMENT",
    "display_basic_aligned",
    "ColorScheme",
    "DefaultColorScheme",
]

_HTML_REPORT_NAMES = {
    "generate_report",
    "render_report_html",
    "ReportMetric",
    "ReportAlignment",
    "ReportSummaryItem",
    "DEFAULT_REPORT_METRICS",
    "DEFAULT_REPORT_SUMMARY_ITEMS",
    "DEFAULT_REPORT_ALIGNMENT",
}

_PYTHON_ALIGNMENT_NAMES = {"display_basic_aligned", "ColorScheme", "DefaultColorScheme"}


def __getattr__(name: str):
    if name in _HTML_REPORT_NAMES:
        from bewer.reporting.html.report import (
            DEFAULT_REPORT_ALIGNMENT,
            DEFAULT_REPORT_METRICS,
            DEFAULT_REPORT_SUMMARY_ITEMS,
            ReportAlignment,
            ReportMetric,
            ReportSummaryItem,
            generate_report,
            render_report_html,
        )

        _attrs = {
            "generate_report": generate_report,
            "render_report_html": render_report_html,
            "ReportMetric": ReportMetric,
            "ReportAlignment": ReportAlignment,
            "ReportSummaryItem": ReportSummaryItem,
            "DEFAULT_REPORT_METRICS": DEFAULT_REPORT_METRICS,
            "DEFAULT_REPORT_SUMMARY_ITEMS": DEFAULT_REPORT_SUMMARY_ITEMS,
            "DEFAULT_REPORT_ALIGNMENT": DEFAULT_REPORT_ALIGNMENT,
        }
        return _attrs[name]
    if name in _PYTHON_ALIGNMENT_NAMES:
        from bewer.reporting.python.alignment import (
            ColorScheme,
            DefaultColorScheme,
            display_basic_aligned,
        )

        _attrs = {
            "display_basic_aligned": display_basic_aligned,
            "ColorScheme": ColorScheme,
            "DefaultColorScheme": DefaultColorScheme,
        }
        return _attrs[name]
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



"""Module for tokenization using regex patterns.

Functions returning regex patterns are generally advised to return uncompiled patterns to allow combination, but
compiled patterns are also allowed, which may be necessary when passing flags to the regex compiler.
"""

from typing import Union

import regex as re

__all__ = ["Tokenizer", "whitespace_pattern", "strip_punctuation_pattern", "strip_punctuation_keep_symbols_pattern"]


def whitespace_pattern() -> re.Pattern:
    """Return a regex pattern that matches sequences of non-whitespace characters.

    Returns:
        re.Pattern: The compiled regex pattern.
    """
    return re.compile(r"\S+")


def strip_punctuation_pattern(split_on_escaped: str | None = None, split_on_pattern: str | None = None) -> re.Pattern:
    """Return a regex pattern that matches tokens without internal whitespace or punctuation per specified characters.

    Args:
        split_on_escaped (str): A string of characters to split on in addition to whitespace. Will be escaped.
        split_on_pattern (str): A regex pattern to split on in addition to whitespace. Will be used as-is.

    Returns:
        re.Pattern: The compiled regex pattern.
    """
    letters_digits = r"\p{L}\p{N}"
    symbols_punctuation_marks = r"\p{S}\p{P}\p{M}"
    if split_on_escaped is None and split_on_pattern is None:
        return re.compile(rf"[{letters_digits}]+([[{symbols_punctuation_marks}]]+[{letters_digits}]+)*", re.V1)
    escaped_split_on = (re.escape(split_on_escaped) if split_on_escaped is not None else "") + (split_on_pattern or "")
    return re.compile(
        rf"[{letters_digits}]+([[{symbols_punctuation_marks}]--[{escaped_split_on}]]+[{letters_digits}]+)*", re.V1
    )


def strip_punctuation_keep_symbols_pattern(
    split_on_escaped: str | None = None,
) -> re.Pattern:
    """
    Return a regex pattern that matches tokens without internal whitespace or punctuation per specified characters, but
    keeps currency symbols, math symbols, and percent signs as separate tokens.

    Args:
        split_on_escaped (str): A string of characters to split on in addition to whitespace. Will be escaped.

    Returns:
        re.Pattern: The compiled regex pattern.
    """
    math_currency = r"\p{Sm}\p{Sc}%"
    strip_pattern = strip_punctuation_pattern(
        split_on_escaped=split_on_escaped,
        split_on_pattern=math_currency,
    )
    return re.compile(rf"([{math_currency}])|({strip_pattern.pattern})", re.V1)


class Tokenizer(object):
    def __init__(
        self,
        pattern: Union[str, re.Pattern],
        name: str = "__no_name__",
    ):
        """Initialize a tokenizer with a regex pattern.

        Args:
            name (str): The config name of the tokenizer.
            pattern (Union[str, re.Pattern]): The regex pattern to use for tokenization.
        """
        self._pattern = re.compile(pattern) if isinstance(pattern, str) else pattern
        self._name = name

    @property
    def pattern(self) -> re.Pattern:
        """Get the regex pattern used by the tokenizer."""
        return self._pattern

    def __call__(self, text: str) -> list[re.Match]:
        """Tokenize the input text using the regex pattern.

        Args:
            text: The input text to tokenize.

        Returns:
            list[re.Match]: A list of regex match objects.
        """
        return list(self._pattern.finditer(text))

    def __repr__(self):
        """Return a string representation of the Tokenizer object."""
        if self._name is None:
            return f"Tokenizer({self._pattern.pattern})"
        return f"Tokenizer({self._name}: {self._pattern.pattern})"



"""Module for styling alignment display with colors.

Alignments can be displayed in the terminal with color coding for different operation types.

In terms of detail level, three main display modes are supported:
- Basic: The tokens are displayed as provided by the alignment. Tokens are separated by a single space.
- Contextualized: The tokens are displayed with surrounding context (new lines may be added/removed for the
                  hypothesis). This mode requires that either the span attributes are set for the operations, or that
                  the alignment token_idx attributes are set and the original texts are accessible.
"""

import shutil
from typing import TYPE_CHECKING

from rich.console import Console
from rich.text import Text

from bewer.alignment.op_type import OpType

if TYPE_CHECKING:
    from bewer.alignment.op import Alignment, Op

__all__ = ["display_basic_aligned", "ColorScheme", "DefaultColorScheme"]


class ColorScheme:
    """Base class for color schemes used in alignment display."""

    PAD: str
    DEL: str
    INS: str
    SUB: str
    MATCH: str


class DefaultColorScheme(ColorScheme):
    PAD = "bright_black"
    DEL = "bright_red"
    INS = "bright_cyan"
    SUB = "bright_yellow"
    MATCH = "default"


def get_padding(length: int, color_scheme: ColorScheme = DefaultColorScheme) -> Text:
    """Get a padding string of spaces of the specified length.

    Args:
        length (int): The length of the padding.

    Returns:
        str: A string of spaces of the specified length.
    """
    return Text(" " * length, style=f"on {color_scheme.PAD}")


def format_match_op(op: "Op", color_scheme: ColorScheme = DefaultColorScheme) -> tuple[Text, Text, int]:
    """Format a match operation for display."""
    len_ref = len(op.ref)
    len_hyp = len(op.hyp)
    length = max(len_ref, len_hyp)
    # NOTE: For matches, the ref and hyp token should be identical, but might differ due to normalization.
    ref_str = Text(op.ref, style=color_scheme.MATCH)
    hyp_str = Text(op.hyp, style=color_scheme.MATCH)
    if len_ref < length:
        ref_str.append(get_padding(length - len_ref, color_scheme=color_scheme))
    if len_hyp < length:
        hyp_str.append(get_padding(length - len_hyp, color_scheme=color_scheme))
    return ref_str, hyp_str, length


def format_substitute_op(op: "Op", color_scheme: ColorScheme = DefaultColorScheme) -> tuple[Text, Text, int]:
    """Format a substitute operation for display."""
    len_ref = len(op.ref)
    len_hyp = len(op.hyp)
    length = max(len_ref, len_hyp)
    ref_str = Text(op.ref, style=color_scheme.SUB)
    hyp_str = Text(op.hyp, style=color_scheme.SUB)
    if len_ref < length:
        ref_str.append(get_padding(length - len_ref, color_scheme=color_scheme))
    if len_hyp < length:
        hyp_str.append(get_padding(length - len_hyp, color_scheme=color_scheme))
    return ref_str, hyp_str, length


def format_insert_op(op: "Op", color_scheme: ColorScheme = DefaultColorScheme) -> tuple[Text, Text, int]:
    """Format an insert operation for display."""
    len_hyp = len(op.hyp)
    hyp_str = Text(op.hyp, style=color_scheme.INS)
    ref_str = get_padding(len_hyp, color_scheme=color_scheme)
    return ref_str, hyp_str, len_hyp


def format_delete_op(op: "Op", color_scheme: ColorScheme = DefaultColorScheme) -> tuple[Text, Text, int]:
    """Format a delete operation for display."""
    len_ref = len(op.ref)
    ref_str = Text(op.ref, style=color_scheme.DEL)
    hyp_str = get_padding(len_ref, color_scheme=color_scheme)
    return ref_str, hyp_str, len_ref


def format_alignment_op(op: "Op", color_scheme: ColorScheme = DefaultColorScheme) -> tuple[Text, Text, int]:
    """Format an alignment operation for display.

    Args:
        op (Op): The alignment operation.

    Returns:
        tuple[str, str, int]: A tuple containing the formatted ref and hyp tokens and their unformatted length.
    """
    if op.type == OpType.MATCH:
        return format_match_op(op, color_scheme=color_scheme)
    if op.type == OpType.SUBSTITUTE:
        return format_substitute_op(op, color_scheme=color_scheme)
    if op.type == OpType.INSERT:
        return format_insert_op(op, color_scheme=color_scheme)
    if op.type == OpType.DELETE:
        return format_delete_op(op, color_scheme=color_scheme)
    raise ValueError(f"Unknown operation type: {op.type}")


def get_line_prefixes(line_number: int) -> tuple[Text, Text]:
    """Get the line prefixes for reference and hypothesis lines."""
    str_num = str(line_number).rjust(4)

    ref_prefix = Text(f"{str_num}  Ref.  ", style="bright_black")
    hyp_prefix = Text("      Hyp.  ", style="bright_black")
    return ref_prefix, hyp_prefix


def display_basic_aligned(
    alignment: "Alignment",
    max_line_length: int | float = 0.5,
    title: str | None = None,
    color_scheme: ColorScheme = DefaultColorScheme,
) -> None:
    """Display the alignment in a basic format.

    Args:
        alignment (Alignment): The alignment to display.
        max_line_length (int | float): The maximum line length for wrapping. If a float, it is interpreted as a
            fraction of the terminal width.
        title (str | None): An optional title to display above the alignment.
        color_scheme (ColorScheme): The color scheme to use for display.

    Prints:
        The formatted alignment to the console.
    """
    if isinstance(max_line_length, float):
        if not (0.0 < max_line_length <= 1.0):
            raise ValueError("If max_line_length is a float, it must be in the range (0.0, 1.0].")
        max_line_length = round(shutil.get_terminal_size().columns * max_line_length)

    # Construct and format lines.
    lines: list[tuple[Text, Text]] = []
    ref_line, hyp_line = get_line_prefixes(line_number=1)
    max_line_length -= len(ref_line)  # Adjust max_line_length to account for prefixes
    for op in alignment:
        ref_str, hyp_str, op_length = format_alignment_op(op, color_scheme=color_scheme)

        if len(ref_line) + op_length > max_line_length:
            lines.append((ref_line, hyp_line))
            ref_line, hyp_line = get_line_prefixes(line_number=len(lines) + 1)

        ref_line.append(ref_str).append(" ")
        hyp_line.append(hyp_str)
        # NOTE: The ref cannot be partial with any of the current alignment algorithms.
        if op.hyp_right_partial:
            hyp_line.append(get_padding(1, color_scheme=color_scheme))
        else:
            hyp_line.append(" ")

    lines.append((ref_line, hyp_line))

    # Build and print output.
    output = Text()
    separator = Text("", style="bright_black")
    if title:
        output.append(Text(title + "\n\n", style="bold bright_black"))  # type: ignore
    for i, (ref, hyp) in enumerate(lines):
        ref.rstrip()
        hyp.rstrip()
        output.append(ref).append("\n").append(hyp).append("\n")
        if i < len(lines) - 1:
            output.append(separator).append("\n")

    Console().print(output)



from __future__ import annotations

from dataclasses import dataclass

from bewer.alignment import OpType
from bewer.metrics.base import METRIC_REGISTRY, ExampleMetric, Metric, MetricParams, metric_value

__all__ = ["KWER"]


class KWER_(ExampleMetric):
    def _get_alignment(self):
        """Get the alignment for the example."""
        return self.example.metrics.levenshtein(
            normalized=self.params.normalized,
            standardizer=self.standardizer,
            tokenizer=self.tokenizer,
            normalizer=self.normalizer,
        ).alignment

    def _get_keyword_matches(self) -> list[slice]:
        return self.example.get_keyword_matches(
            vocab=self.params.vocab,
            normalized=self.params.normalized,
            allow_subsets=self.params.allow_subsets,
        )

    @metric_value
    def num_errors(self) -> int:
        """Get the number of keywords incorrectly transcribed in the hypothesis text."""
        keyword_matches = self._get_keyword_matches()
        if not keyword_matches:
            return 0
        alignment = self._get_alignment()
        num_errors = 0
        for keyword_match in keyword_matches:
            for ref_index in range(keyword_match.start, keyword_match.stop):
                op_index = alignment.ref_index_mapping.get(ref_index)
                if alignment[op_index].type != OpType.MATCH:
                    num_errors += 1
                    break

        return num_errors

    @metric_value
    def num_keywords(self) -> int:
        """Get the number of keywords in the reference text."""
        return len(self._get_keyword_matches())

    @metric_value(main=True)
    def value(self) -> float:
        """Get the example-level keyword error rate."""
        if self.num_keywords == 0:
            return float(self.num_errors)
        return self.num_errors / self.num_keywords


@METRIC_REGISTRY.register("kwer")
class KWER(Metric):
    short_name_base = "KWER"
    long_name_base = "Keyword Error Rate"
    description = (
        "Keyword error rate (KWER) is computed as the number of keywords (or key terms) incorrectly transcribed in "
        "the hypothesis texts, divided by the total number of keywords identified in the reference texts. A keyword "
        "may consist of one or more tokens, but is treated as a single unit for the purpose of KWER calculation."
    )
    example_cls = KWER_

    @dataclass
    class param_schema(MetricParams):
        """Parameters for the KWER metric.

        Attributes:
            vocab: The vocabulary name to use for keyword identification.
            normalized: Whether to use normalized tokens for alignment and keyword matching.
            allow_subsets: Whether to allow subset matches. If False, overlapping keyword matches
                are deduplicated, keeping only the longest match.
        """

        vocab: str
        normalized: bool = True
        allow_subsets: bool = True

        def validate(self) -> None:
            """Validate that the metric can be computed with the given parameters and source data."""
            is_dynamic_vocab = self.vocab in self.metric.dataset._dynamic_keyword_vocabs
            is_static_vocab = self.vocab in self.metric.dataset._static_keyword_vocabs
            if not is_dynamic_vocab and not is_static_vocab:
                raise ValueError(f"Vocabulary '{self.vocab}' not found in dataset keyword vocabularies.")

    @metric_value
    def num_errors(self) -> int:
        """Get the number of keywords incorrectly transcribed in the hypothesis texts."""
        return sum([self.get_example_metric(example).num_errors for example in self._src])

    @metric_value
    def num_keywords(self) -> int:
        """Get the number of keywords in the reference texts."""
        return sum([self.get_example_metric(example).num_keywords for example in self._src])

    @metric_value(main=True)
    def value(self) -> float:
        """Get the keyword error rate."""
        if self.num_keywords == 0:
            return float(self.num_errors)
        return self.num_errors / self.num_keywords



from __future__ import annotations

from bewer.metrics.base import METRIC_REGISTRY, ExampleMetric, Metric, metric_value

__all__ = ["DatasetSummary"]


class DatasetSummary_(ExampleMetric):
    @metric_value
    def num_ref_words(self) -> int:
        """The number of tokens in the reference text."""
        return len(self.example.ref.tokens)

    @metric_value
    def num_ref_chars(self) -> int:
        """The number of characters in the reference text."""
        return len(self.example.ref.standardized)

    @metric_value
    def num_hyp_words(self) -> int:
        """The number of tokens in the hypothesis text."""
        return len(self.example.hyp.tokens)

    @metric_value
    def num_hyp_chars(self) -> int:
        """The number of characters in the hypothesis text."""
        return len(self.example.hyp.standardized)


@METRIC_REGISTRY.register("summary")
class DatasetSummary(Metric):
    short_name_base = "Summary"
    long_name_base = "Dataset Summary"
    description = (
        "Dataset Summary provides basic statistics about the dataset, including the number of words and characters "
        "in both the reference and hypothesis texts."
    )
    example_cls = DatasetSummary_

    @metric_value
    def num_examples(self) -> int:
        """The total number of examples in the dataset."""
        return len(self._src)

    @metric_value
    def num_ref_words(self) -> int:
        """The total number of tokens in the reference texts."""
        return sum(self.get_example_metric(example).num_ref_words for example in self._src)

    @metric_value
    def num_ref_chars(self) -> int:
        """The total number of characters in the reference texts."""
        return sum(self.get_example_metric(example).num_ref_chars for example in self._src)

    @metric_value
    def num_hyp_words(self) -> int:
        """The total number of tokens in the hypothesis texts."""
        return sum(self.get_example_metric(example).num_hyp_words for example in self._src)

    @metric_value
    def num_hyp_chars(self) -> int:
        """The total number of characters in the hypothesis texts."""
        return sum(self.get_example_metric(example).num_hyp_chars for example in self._src)



from rich.console import Console
from rich.table import Table

__all__: list[str] = []  # Internal


def print_metric_table(
    rows: list[tuple[tuple[str, str, str], tuple[str, str, str] | None]],
) -> None:
    """
    Print a table of metric values.

    Args:
        rows (list[tuple[tuple[str, str, str], tuple[str, str, str] | None]]): The rows of the table. Two rows per
            metric: one for the main metric and one for the example metric (if any). Each triple contains the name,
            main value, and other values as a single comma-separated string.

    Prints:
        A table of metric values.
    """
    table = Table(title="Registered metrics")
    table.add_column("Name", justify="left", style="bright_cyan", no_wrap=True)
    table.add_column("Level", justify="left", style="bright_black", no_wrap=True)
    table.add_column("Main", style="bright_magenta")
    table.add_column("Other", justify="left", style="bright_black")

    for metric_name, (main_row, example_row) in rows:
        end_section = True if example_row is None else False
        table.add_row(metric_name, "dataset", *main_row, end_section=end_section)
        if example_row is not None:
            table.add_row("", "example", *example_row, end_section=True)

    console = Console()
    console.print(table)



__all__ = [
    "generate_report",
    "render_report_html",
    "ReportMetric",
    "ReportAlignment",
    "ReportSummaryItem",
    "DEFAULT_REPORT_METRICS",
    "DEFAULT_REPORT_SUMMARY_ITEMS",
    "DEFAULT_REPORT_ALIGNMENT",
    "HTMLAlignmentColors",
    "HTMLDefaultAlignmentColors",
    "HTMLBaseColors",
    "HTMLAlignmentLabels",
]

_REPORT_NAMES = {
    "generate_report",
    "render_report_html",
    "ReportMetric",
    "ReportAlignment",
    "ReportSummaryItem",
    "DEFAULT_REPORT_METRICS",
    "DEFAULT_REPORT_SUMMARY_ITEMS",
    "DEFAULT_REPORT_ALIGNMENT",
}

_COLOR_SCHEME_NAMES = {"HTMLAlignmentColors", "HTMLDefaultAlignmentColors", "HTMLBaseColors"}


def __getattr__(name: str):
    if name in _REPORT_NAMES:
        from bewer.reporting.html import report

        return getattr(report, name)
    if name in _COLOR_SCHEME_NAMES:
        from bewer.reporting.html import color_schemes

        return getattr(color_schemes, name)
    if name == "HTMLAlignmentLabels":
        from bewer.reporting.html.labels import HTMLAlignmentLabels

        return HTMLAlignmentLabels
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Optional, Union

import ahocorasick

from bewer.core.text import Text, TextType, TokenList
from bewer.preprocessing.context import NORMALIZER_NAME, STANDARDIZER_NAME, TOKENIZER_NAME

if TYPE_CHECKING:
    from bewer.core.dataset import Dataset
    from bewer.core.example import Example

__all__ = ["Keyword", "KeywordNotFoundWarning"]


class KeywordNotFoundWarning(UserWarning):
    pass


warnings.filterwarnings("always", category=KeywordNotFoundWarning)


class Keyword(Text):
    """A keyword that can locate itself within reference text tokens.

    Inherits standardized, tokens, and pipeline caching from Text.
    Adds contiguous token matching against a reference TokenList.
    """

    def __init__(
        self,
        raw: str,
        src: Optional[Union["Example", "Dataset"]] = None,
    ):
        super().__init__(raw=raw, src=src, text_type=TextType.KEYWORD)

    def __repr__(self):
        text = self.raw if len(self.raw) <= 46 else self.raw[:46] + "..."
        return f'Keyword("{text}")'


class KeywordTrie:
    """Aho-Corasick automaton for efficient multi-keyword matching in token sequences."""

    def __init__(
        self,
        keywords: set[Keyword],
        normalized: bool = True,
        add_capitalized: bool = False,
    ):
        """Initialize the automaton with the given set of keywords.

        Args:
            keywords: A set of Keyword objects to build the automaton from.
            normalized: Whether to use normalized text for matching. If False, uses raw text.
            add_capitalized: Whether to add capitalized versions of keywords for case-insensitive matching.
                Only applies if normalized is False and is only applied to the first word in an n-gram keyword.
        """
        self.normalized = normalized
        self.add_capitalized = add_capitalized

        # Collect token patterns from keywords, tracking origin
        patterns = []
        self._pattern_keywords: dict[tuple[int, ...], set[str]] = {}
        keyword_patterns = []
        for keyword in keywords:
            tokens = keyword.tokens.normalized if normalized else keyword.tokens.raw
            token_pattern = tuple(tokens)
            if not token_pattern:
                continue
            keyword_patterns.append((keyword.raw, token_pattern))
            patterns.append(token_pattern)

        # Handle capitalization variants
        if add_capitalized and not normalized:
            for _, p in keyword_patterns:
                first_cap = p[0].capitalize()
                if first_cap != p[0]:
                    patterns.append((first_cap,) + p[1:])

        # Build vocab: token string -> int for KEY_SEQUENCE mode
        self._vocab = {w: i for i, w in enumerate({w for p in patterns for w in p})}
        self._unknown = len(self._vocab)

        # Map int_pattern -> set of keyword raw strings (for warn_missing)
        for kw_raw, token_pattern in keyword_patterns:
            int_pattern = tuple(self._vocab[w] for w in token_pattern)
            self._pattern_keywords.setdefault(int_pattern, set()).add(kw_raw)

        # Build Aho-Corasick automaton
        self._automaton = ahocorasick.Automaton(ahocorasick.STORE_ANY, ahocorasick.KEY_SEQUENCE)
        seen = set()
        for pattern in patterns:
            int_pattern = tuple(self._vocab[w] for w in pattern)
            if int_pattern not in seen:
                self._automaton.add_word(int_pattern, len(pattern))
                seen.add(int_pattern)
        self._automaton.make_automaton()

    def find_in_tokens(
        self,
        tokens: TokenList,
        allow_subsets: bool = True,
        warn_missing: bool = False,
    ) -> list[slice]:
        """Find all contiguous token sequences that match any keyword in the automaton.

        Args:
            tokens: The token list to search in.
            allow_subsets: Whether to allow subset matches (default True).
            warn_missing: If True, emit a KeywordNotFoundWarning for each keyword whose
                token pattern was not found in the text. Uses tokens.src to identify the example.
        """
        token_strings = tokens.normalized if self.normalized else tokens.raw
        int_text = tuple(self._vocab.get(w, self._unknown) for w in token_strings)

        matches = []
        found_keywords: set[str] | None = set() if warn_missing else None
        for end_idx, pattern_len in self._automaton.iter(int_text):
            start_idx = end_idx - pattern_len + 1
            matches.append(slice(start_idx, end_idx + 1))
            if found_keywords is not None:
                int_pattern = int_text[start_idx : end_idx + 1]
                for kw_raw in self._pattern_keywords.get(int_pattern, ()):
                    found_keywords.add(kw_raw)

        if warn_missing:
            example_index = getattr(getattr(tokens.src, "src", None), "index", None)
            for kw_raws in self._pattern_keywords.values():
                for kw_raw in kw_raws:
                    if kw_raw not in found_keywords:
                        warnings.warn(
                            f"Keyword '{kw_raw}' not found in reference tokens: Example {example_index}.",
                            KeywordNotFoundWarning,
                        )

        if not allow_subsets:
            matches = _remove_subset_matches(matches)

        return matches


def _remove_duplicate_matches(matches: list[slice]) -> list[slice]:
    """Remove exact duplicate matches, preserving order."""
    seen: set[tuple[int, int]] = set()
    result = []
    for m in matches:
        key = (m.start, m.stop)
        if key not in seen:
            seen.add(key)
            result.append(m)
    return result


def _remove_subset_matches(matches: list[slice]) -> list[slice]:
    """Remove matches that are subsets of other matches, preferring longer matches."""
    if not matches:
        return matches
    # Sort by start ascending, then by length descending
    matches.sort(key=lambda s: (s.start, s.start - s.stop))
    result = [matches[0]]
    for m in matches[1:]:
        prev = result[-1]
        # Skip if fully contained within previous match
        if m.start >= prev.start and m.stop <= prev.stop:
            continue
        result.append(m)
    return result


def get_keyword_trie(
    vocabs: dict[str, set[Keyword]],
    cache: dict[tuple, Optional[KeywordTrie]],
    vocab: str,
    normalized: bool = True,
    add_capitalized: bool = False,
) -> Optional[KeywordTrie]:
    """Get or build a trie for the keywords in the specified vocabulary."""
    trie_key = (
        STANDARDIZER_NAME.get(),
        TOKENIZER_NAME.get(),
        NORMALIZER_NAME.get() if normalized else None,
        add_capitalized,
        vocab,
    )
    if trie_key in cache:
        return cache[trie_key]

    keywords = vocabs.get(vocab, None)
    if keywords is None:
        return None

    trie = KeywordTrie(
        keywords,
        normalized=normalized,
        add_capitalized=add_capitalized,
    )
    cache[trie_key] = trie
    return trie



from typing import TYPE_CHECKING, Optional

from bewer.core.keyword import Keyword, KeywordTrie, _remove_duplicate_matches, _remove_subset_matches, get_keyword_trie
from bewer.core.text import Text, TextType
from bewer.metrics.base import ExampleMetricCollection
from bewer.preprocessing.context import NORMALIZER_NAME, STANDARDIZER_NAME, TOKENIZER_NAME

if TYPE_CHECKING:
    from bewer.core.dataset import Dataset

__all__ = ["Example"]


class Example:
    """
    BeWER example representation.

    Attributes:
        ref (Text): Reference text object.
        hyp (Text): Hypothesis text object.
        keywords (dict[str, set[Keyword]]): Keywords grouped by vocabulary name.
        metrics (ExampleMetricCollection): Metrics collection for this example.
    """

    def __init__(
        self,
        ref: str,
        hyp: str,
        keywords: dict[str, list[str]] | None = None,
        src: Optional["Dataset"] = None,
        index: Optional[int] = None,
    ):
        """
        Initialize the Example object.

        Args:
            ref: Reference text.
            hyp: Hypothesis text.
            keywords: Keywords associated with the example. Missing terms are retained; warnings are emitted
                during keyword trie matching if a term cannot be matched in the reference tokens.
            src: Parent Dataset object. Can be set later via set_source().
            index: The index of the example in the dataset.
        """
        self._index = index
        self._cache_keyword_matches = {}
        self._cache_keyword_tries = {}
        self._pipelines = None

        self._src = None
        if src is not None:
            self.set_source(src)

        self.metrics = ExampleMetricCollection(self)
        self.ref = Text(ref, src=self, text_type=TextType.REF)
        self.hyp = Text(hyp, src=self, text_type=TextType.HYP)
        self.keywords = self._prepare_keywords(keywords)

    @property
    def index(self) -> Optional[int]:
        """Get the example index."""
        return self._index

    @property
    def src(self) -> Optional["Dataset"]:
        """Get the parent Dataset object."""
        return self._src

    @property
    def pipelines(self):
        return self._pipelines

    @property
    def vocabs(self) -> set[str]:
        """Get the set of all keyword vocabularies associated with this example."""
        vocabs = set(self.keywords.keys())
        if self._src is not None:
            vocabs.update(self._src._dynamic_keyword_vocabs.keys())
        return vocabs

    def set_source(self, src: "Dataset") -> None:
        """Set the parent Dataset object.

        Args:
            src: The parent Dataset object.

        Raises:
            ValueError: If source is already set.
        """
        if self._src is not None:
            raise ValueError("Source already set for Example")
        self._src = src
        self._pipelines = src.pipelines

    def _prepare_keywords(self, keywords: dict[str, set[str]] | None) -> dict[str, set[Keyword]]:
        """Prepare keywords dictionary by converting keywords to Keyword objects."""
        if keywords is None:
            return {}

        prepared_keywords = {}
        for vocab_name, vocab_keywords in keywords.items():
            if len(vocab_keywords) == 0:
                continue
            prepared_keywords[vocab_name] = set(Keyword(keyword, src=self) for keyword in vocab_keywords)

        return prepared_keywords

    def _get_keyword_trie(
        self,
        vocab: str,
        normalized: bool = True,
        add_capitalized: bool = False,
    ) -> Optional["KeywordTrie"]:
        """Get or build a trie for the keywords in the specified vocabulary."""
        return get_keyword_trie(
            self.keywords,
            self._cache_keyword_tries,
            vocab,
            normalized=normalized,
            add_capitalized=add_capitalized,
        )

    def get_keyword_matches(
        self,
        vocab: str,
        normalized: bool = True,
        add_capitalized: bool = False,
        allow_subsets: bool = True,
    ) -> list[slice]:
        if vocab not in self.keywords and vocab not in self.src._dynamic_keyword_vocabs:
            return []

        cache_key = (
            STANDARDIZER_NAME.get(),
            TOKENIZER_NAME.get(),
            NORMALIZER_NAME.get() if normalized else None,
            add_capitalized,
            allow_subsets,
            vocab,
        )
        if cache_key in self._cache_keyword_matches:
            return self._cache_keyword_matches[cache_key]

        matches = []
        if vocab in self.keywords:
            example_trie = self._get_keyword_trie(vocab, normalized=normalized, add_capitalized=add_capitalized)
            matches += example_trie.find_in_tokens(self.ref.tokens, warn_missing=True)

        if vocab in self.src._dynamic_keyword_vocabs:
            dataset_trie = self.src._get_keyword_trie(vocab, normalized=normalized, add_capitalized=add_capitalized)
            if dataset_trie is not None:
                matches += dataset_trie.find_in_tokens(self.ref.tokens)

        if matches:
            if allow_subsets:
                matches = _remove_duplicate_matches(matches)
            else:
                matches = _remove_subset_matches(matches)

        self._cache_keyword_matches[cache_key] = matches
        return matches

    def __hash__(self):
        return hash((self.ref, self.hyp, self._index))

    def __repr__(self):
        ref = self.ref.raw if len(self.ref.raw) <= 45 else self.ref.raw[:42] + "..."
        hyp = self.hyp.raw if len(self.hyp.raw) <= 45 else self.hyp.raw[:42] + "..."
        return f'Example(ref="{ref}", hyp="{hyp}")'



from typing import TYPE_CHECKING, Optional

import regex as re

from bewer.core.caching import pipeline_cached_property
from bewer.preprocessing.context import NORMALIZER_NAME
from bewer.reporting.python.utils import highlight_span

if TYPE_CHECKING:
    from bewer.core.text import Text

__all__ = ["Token"]


class Token:
    """BeWER Token representation.

    Attributes:
        raw (str): The raw string of the token.
        start (int): The starting index of the token in the text.
        end (int): The ending index of the token in the text.
        index (int | None): The index of the token in the token list.
        slice (slice): A slice object representing the token's position in the text.
        normalized (str | None): The normalized string of the token.
    """

    def __init__(
        self,
        raw: str,
        start: int,
        end: int,
        index: Optional[int] = None,
        src: Optional["Text"] = None,
    ):
        """Initialize Token.

        Args:
            raw: The raw token string.
            start: Starting character index in the source text.
            end: Ending character index in the source text.
            index: Token index in the token list.
            src: Parent Text object. Can be set later via set_source().
        """
        self._raw = raw
        self.start = start
        self.end = end
        self.index = index
        self.slice = slice(self.start, self.end)

        self._cache_normalized = {}

        self._src = None
        self._pipelines = None
        if src is not None:
            self.set_source(src)

    @property
    def src(self) -> Optional["Text"]:
        """Get the parent Text object."""
        return self._src

    def set_source(self, src: "Text") -> None:
        """Set the parent Text object.

        Args:
            src: The parent Text object.

        Raises:
            ValueError: If source is already set.
        """
        if self._src is not None:
            raise ValueError("Source already set for Token")

        self._src = src
        self._pipelines = src.pipelines

    @property
    def raw(self) -> str:
        """The raw string of the token as extracted during tokenization."""
        return self._raw

    @pipeline_cached_property(NORMALIZER_NAME)
    def normalized(self, normalizer):
        """The normalized string of the token after applying the active normalizer."""
        return normalizer(self.raw)

    def inctx(self, width: int = 20, highlight: bool = False, add_ellipsis: bool = True) -> str:
        """Get the context of the token in the source text.

        Args:
            width (int): The number of characters of context to show on each side.
            highlight (bool): Whether to highlight the token span in the context.
            add_ellipsis (bool): Whether to add ellipsis around the context, if starting/ending not within width.

        Returns:
            str: The context string.
        """
        if self._src is None:
            raise ValueError("Source text is not set. Cannot get context.")
        start = max(0, self.start - width)
        end = min(len(self._src.raw), self.end + width)
        ctx_span = self._src.raw[start:end]
        if highlight:
            ctx_span = highlight_span(ctx_span, self.start - start, self.end - start, "bold green")
        if add_ellipsis:
            start_marker = "..." if self.start - width > 0 else ""
            end_marker = "..." if self.end + width < len(self._src.raw) else ""
            ctx_span = start_marker + ctx_span + end_marker
        return ctx_span

    @classmethod
    def from_match(
        cls,
        match: re.Match,
        index: int,
        src: Optional["Text"] = None,
    ) -> "Token":
        """
        Create a Token object from a regex match object.

        Args:
            match (re.Match): The regex match object.
            index (int): Token index in the token list.
            src (Text): Parent Text object.

        Returns:
            Token: The created Token object.
        """
        return cls(
            raw=match.group(),
            start=match.start(),
            end=match.end(),
            index=index,
            src=src,
        )

    def __eq__(self, other):
        if not isinstance(other, Token):
            return False
        return self.start == other.start and self.end == other.end and self.raw == other.raw

    def __repr__(self):
        return f'Token("{self.raw}")'



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property

from bewer.metrics.base import METRIC_REGISTRY, ExampleMetric, Metric, MetricParams, metric_value

__all__ = ["MTR"]


class MTR_(ExampleMetric):
    @metric_value
    def num_matches(self) -> int:
        """Get the number of medical keywords correctly transcribed in the hypothesis text."""
        return self.num_keywords - self.parent_metric._kwer_metric.get_example_metric(self.example).num_errors

    @metric_value
    def num_keywords(self) -> int:
        """Get the number of medical keywords in the reference text."""
        return self.parent_metric._kwer_metric.get_example_metric(self.example).num_keywords

    @metric_value(main=True)
    def value(self) -> float:
        """Get the example-level medical term recall."""
        return 1 - self.parent_metric._kwer_metric.get_example_metric(self.example).value


@METRIC_REGISTRY.register("mtr")
class MTR(Metric):
    short_name_base = "MTR"
    long_name_base = "Medical Term Recall"
    description = (
        "Medical term recall (MTR) is computed as the number of medical keywords (or key terms) correctly identified "
        "in the hypothesis texts, divided by the total number of medical keywords identified in the reference texts. "
        "A keyword may consist of one or more tokens, but is treated as a single unit for the purpose of MTR "
        "calculation. MTR is the complement of KWER (Keyword Error Rate) with the keyword vocabulary set to "
        "'medical_terms'. Specifically, MTR can be calculated as 1 - KWER."
    )
    example_cls = MTR_

    @dataclass
    class param_schema(MetricParams):
        """Parameters for the MTR metric.

        Attributes:
            normalized: Whether to use normalized tokens for alignment and keyword matching.
        """

        normalized: bool = True

        def validate(self) -> None:
            """Validate that the metric can be computed with the given parameters and source data."""
            is_dynamic_vocab = "medical_terms" in self.metric.dataset._dynamic_keyword_vocabs
            is_static_vocab = "medical_terms" in self.metric.dataset._static_keyword_vocabs
            if not is_dynamic_vocab and not is_static_vocab:
                raise ValueError("Vocabulary 'medical_terms' not found in dataset keyword vocabularies.")

    @cached_property
    def _kwer_metric(self):
        """Get the corresponding KWER metric instance for calculating MTR."""
        return self.dataset.metrics.kwer(
            vocab="medical_terms",
            normalized=self.params.normalized,
            standardizer=self.standardizer,
            tokenizer=self.tokenizer,
            normalizer=self.normalizer,
        )

    @metric_value
    def num_matches(self) -> int:
        """Get the number of medical keywords correctly transcribed in the hypothesis texts."""
        return self._kwer_metric.num_keywords - self._kwer_metric.num_errors

    @metric_value
    def num_keywords(self) -> int:
        """Get the number of medical keywords in the reference texts."""
        return self._kwer_metric.num_keywords

    @metric_value(main=True)
    def value(self) -> float:
        """Get the medical term recall."""
        return 1 - self._kwer_metric.value



import ast
from functools import cached_property
from importlib import resources
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, Optional, Union

import pandas as pd
from omegaconf import OmegaConf

from bewer.configs.resolve import resolve_pipelines
from bewer.core.example import Example
from bewer.core.keyword import Keyword, get_keyword_trie
from bewer.core.text import TokenList
from bewer.metrics.base import MetricCollection

__all__ = ["Dataset", "TextList", "TextTokenList"]

if TYPE_CHECKING:
    from bewer.core.keyword import KeywordTrie
    from bewer.core.text import Text


def _is_list_literal(s):
    try:
        return isinstance(ast.literal_eval(s), list)
    except (ValueError, SyntaxError):
        return False


class Dataset(object):
    """BeWER dataset.

    Attributes:
        config (OmegaConf): The resolved configuration object.
        pipelines: The resolved preprocessing pipelines.
        examples (list[Example]): A list of Example objects.
        metrics (MetricCollection): A metrics collection for the dataset.
        refs (TextList): The reference texts in the dataset.
        hyps (TextList): The hypothesis texts in the dataset.
    """

    def __init__(self, config: str | None = None):
        """Initialize the Dataset.

        The dataset must be populated using one of the load_* methods or manually using the add() method.

        Args:
            config (str | None): Path to the configuration file. If None, uses the default configuration.
        """
        self.config_path = self.get_config_path(config)
        self.config = OmegaConf.load(self.config_path)
        self._pipelines = resolve_pipelines(self.config)
        self.examples = []
        self._dynamic_keyword_vocabs = {}
        self._static_keyword_vocabs = {}
        self._cache_keyword_tries = {}
        self.metrics = MetricCollection(self)

    @property
    def pipelines(self):
        return self._pipelines

    @cached_property
    def refs(self) -> "TextList":
        """Get the reference texts as a TextList object.

        Returns:
            TextList: The reference texts.
        """
        return TextList([example.ref for example in self.examples])

    @cached_property
    def hyps(self) -> "TextList":
        """Get the hypothesis texts as a TextList object.

        Returns:
            TextList: The hypothesis texts.
        """
        return TextList([example.hyp for example in self.examples])

    def add(self, ref: str, hyp: str, keywords: dict[str, list[str]] | None = None) -> None:
        """Add an example to the dataset."""
        if keywords is not None:
            keywords = {name: set(kw_list) for name, kw_list in keywords.items()}
            for name, kw_set in keywords.items():
                self._update_static_keyword_vocab(name, kw_set)
        example = Example(ref, hyp, keywords=keywords, src=self, index=len(self))
        self.examples.append(example)

    def load_dataset(self, dataset, ref_col="ref", hyp_col="hyp", keyword_cols: list | None = None) -> None:
        """Load a Hugging Face dataset."""
        raise NotImplementedError("load_dataset() method not implemented.")

    def load_pandas(self, df: pd.DataFrame, ref_col="ref", hyp_col="hyp", keyword_cols: list | None = None) -> None:
        """Add a pandas DataFrame to the dataset."""
        if not isinstance(df, pd.DataFrame):
            raise TypeError("df must be a pandas DataFrame")
        if keyword_cols is None:
            keyword_cols = []

        for col in keyword_cols:
            df[col] = self._infer_keyword_column(df[col])

        # Add examples to the dataset
        for row in df.itertuples(index=False):
            hyp = getattr(row, hyp_col)
            ref = getattr(row, ref_col)
            if len(keyword_cols) > 0:
                keywords = {}
                for col in keyword_cols:
                    keywords[col] = getattr(row, col)
            else:
                keywords = None
            self.add(ref, hyp, keywords=keywords)

    def load_csv(self, csv_file: str, ref_col="ref", hyp_col="hyp", keyword_cols: list | None = None, **kwargs) -> None:
        """Add a CSV file to the dataset."""
        df = pd.read_csv(csv_file, **kwargs)
        self.load_pandas(df, ref_col, hyp_col, keyword_cols)

    def load_jsonl(
        self, jsonl_file: str, ref_col="ref", hyp_col="hyp", keyword_cols: list | None = None, **kwargs
    ) -> None:
        """Add a JSONL file to the dataset."""
        df = pd.read_json(jsonl_file, lines=True, **kwargs)
        self.load_pandas(df, ref_col, hyp_col, keyword_cols)

    def add_keyword_list(self, name: str, keywords: Iterable[str]) -> None:
        """Add a named keyword vocabulary to the dataset.

        Keywords are matched against the reference text of each example (including already added examples).

        Args:
            name (str): The name of the keyword vocabulary.
            keywords (Iterable[str]): The keywords to add.
        """
        if not isinstance(keywords, Iterable) or isinstance(keywords, str):
            raise TypeError("keywords must be an iterable of strings")

        keywords = set(keywords)

        for keyword in keywords:
            if not isinstance(keyword, str):
                raise TypeError(f"keywords must be an iterable of strings, but got element of type {type(keyword)}")

        self._update_dynamic_keyword_vocab(name, keywords)

    def add_keyword_file(self, name: str, keyword_file: str) -> None:
        """Add a named keyword vocabulary to the dataset from a file.

        The file should be plain text and contain one keyword per line.

        Keywords are matched against the reference text of each example (including already added examples).

        Args:
            name (str): The name of the keyword vocabulary.
            keyword_file (str): Path to the keyword file.
        """
        if not Path(keyword_file).is_file():
            raise FileNotFoundError(f"Keyword file {keyword_file} not found")

        with open(keyword_file, "r") as f:
            keywords = f.read().strip().splitlines()

        self.add_keyword_list(name, keywords)

    def _get_keyword_trie(
        self, vocab: str, normalized: bool = True, add_capitalized: bool = False
    ) -> Optional["KeywordTrie"]:
        """Get a trie for the specified keyword vocabulary."""
        return get_keyword_trie(
            self._dynamic_keyword_vocabs,
            self._cache_keyword_tries,
            vocab,
            normalized=normalized,
            add_capitalized=add_capitalized,
        )

    @staticmethod
    def get_config_path(config_path: str | None) -> str:
        """Get the configuration path."""
        if config_path is None or not Path(config_path).is_file():
            config_path = "base" if config_path is None else config_path
            return resources.files("bewer.configs").joinpath(f"{config_path}.yml")
        return Path(config_path).resolve()

    def _infer_keyword_column(self, series: pd.Series) -> pd.Series:
        """Infer the keyword terms from a pandas Series."""
        if series.map(_is_list_literal).all():
            series = series.apply(ast.literal_eval)
            return series
        elif series.map(lambda x: isinstance(x, str)).all():
            series = series.apply(lambda x: [x])
            return series
        elif series.map(lambda x: isinstance(x, list)).all():
            return series
        else:
            raise ValueError(f"Column {series.name} is not a list (or literal) or string")

    def _update_dynamic_keyword_vocab(self, name: str, keywords: set[str]) -> None:
        """Update the keyword vocabulary with new keywords."""
        keywords = set(Keyword(keyword, src=self) for keyword in keywords)
        if name in self._dynamic_keyword_vocabs:
            self._dynamic_keyword_vocabs[name].update(keywords)
        else:
            self._dynamic_keyword_vocabs[name] = set(keywords)

    def _update_static_keyword_vocab(self, name: str, keywords: set[str]) -> None:
        """Update the static keyword vocabulary with new keywords."""
        keywords = set(Keyword(keyword, src=self) for keyword in keywords)
        if name in self._static_keyword_vocabs:
            self._static_keyword_vocabs[name].update(keywords)
        else:
            self._static_keyword_vocabs[name] = set(keywords)

    def __len__(self) -> int:
        """Get the number of examples in the dataset."""
        return len(self.examples)

    def __getitem__(self, index: int) -> Example:
        """Get an example by index."""
        return self.examples[index]

    def __iter__(self):
        """Iterate over the examples in the dataset."""
        return iter(self.examples)

    def __repr__(self):
        """Get a string representation of the dataset."""
        return f"Dataset({len(self.examples)} examples)"


class TextList(tuple["Text", ...]):
    """An immutable sequence of Text objects."""

    def __new__(cls, iterable=()):
        return super().__new__(cls, iterable)

    @property
    def raw(self) -> list[str]:
        """Get the raw texts as a regular Python list.

        Returns:
            list[str]: The raw text.
        """
        return [text.raw for text in self]

    @property
    def standardized(self) -> list[str]:
        """Get the standardized texts as a regular Python list.

        Returns:
            list[str]: The standardized texts.
        """
        return [text.standardized for text in self]

    @property
    def tokens(self) -> "TextTokenList":
        """Get the tokens as a TextTokenList object.

        Returns:
            TextTokenList: The tokens.
        """
        return TextTokenList([text.tokens for text in self])

    def __getitem__(self, index: int | slice) -> Union["Text", "TextList"]:
        if isinstance(index, slice):
            return TextList(super().__getitem__(index))
        return super().__getitem__(index)

    def __add__(self, other: "TextList") -> "TextList":
        return TextList(super().__add__(other))

    def __repr__(self):
        texts = self[:60]
        texts_str = ",\n ".join([repr(text) for text in texts])
        if len(self) > 60:
            texts_str += ",\n ..."
        return f"TextList([\n {texts_str}]\n)"


class TextTokenList(tuple["TokenList", ...]):
    """An immutable sequence of TokenList objects."""

    def __new__(cls, iterable=()):
        return super().__new__(cls, iterable)

    @property
    def raw(self) -> list[list[str]]:
        """Get the raw tokens as a regular Python list.

        Returns:
            list[list[str]]: The raw tokens.
        """
        return [tokens.raw for tokens in self]

    @property
    def normalized(self) -> list[list[str]]:
        """Get the normalized tokens as a regular Python list.

        Returns:
            list[list[str]]: The normalized tokens.
        """
        return [tokens.normalized for tokens in self]

    @property
    def flat(self) -> "TokenList":
        """Flatten the TextTokenList into a TokenList.

        Returns:
            TokenList: The flattened TokenList.
        """
        return TokenList(chain(*self))

    def __getitem__(self, index: int | slice) -> Union["TokenList", "TextTokenList"]:
        if isinstance(index, slice):
            return TextTokenList(super().__getitem__(index))
        return super().__getitem__(index)

    def __add__(self, other: "TextTokenList") -> "TextTokenList":
        return TextTokenList(super().__add__(other))

    def __repr__(self):
        text_tokens = self[:60]
        text_tokens_str = ",\n ".join([tokens._sub_repr() for tokens in text_tokens])
        if len(self) > 60:
            text_tokens_str += ",\n ..."
        return f"TextTokenList([\n {text_tokens_str}]\n)"



"""Context-sensitive caching descriptor for pipeline-derived properties."""

from contextvars import ContextVar
from typing import Any, Callable

from bewer.preprocessing.context import PIPELINE_STAGES

__all__ = ["pipeline_cached_property"]

# Build lookup: context_var -> (context_vars_up_to_and_including, pipeline_attr)
_STAGE_INDEX = {}
for i, (cv, attr) in enumerate(PIPELINE_STAGES):
    _STAGE_INDEX[cv] = (tuple(cv for cv, _ in PIPELINE_STAGES[: i + 1]), attr)


class PipelineCachedProperty:
    """Descriptor that implements context-sensitive lazy caching for pipeline properties.

    Reads one or more ContextVars to form a cache key, looks up the pipeline function
    from ``instance._pipelines``, calls it on cache miss, and stores the result in a
    per-instance dict whose name is auto-derived as ``_cache_{property_name}``.

    Use via the :func:`pipeline_cached_property` decorator factory.
    """

    def __init__(
        self,
        context_vars: tuple[ContextVar, ...],
        pipeline: str,
        compute: Callable[[Any, Any], Any],
        doc: str | None = None,
    ):
        self._context_vars = context_vars
        self._pipeline = pipeline
        self._compute = compute
        self.__doc__ = doc

    def __set_name__(self, owner: type, name: str) -> None:
        self._name = name
        self._cache_attr = f"_cache_{name}"

    def __get__(self, instance: Any, owner: type | None = None) -> Any:
        if instance is None:
            return self

        # Build cache key from active context variable values
        key_parts = tuple(cv.get() for cv in self._context_vars)
        cache_key = key_parts[0] if len(key_parts) == 1 else key_parts

        # Check per-instance cache
        cache = getattr(instance, self._cache_attr)
        if cache_key in cache:
            return cache[cache_key]

        # Resolve pipeline function (last context var names the pipeline entry)
        pipelines = instance._pipelines
        if pipelines is None:
            raise ValueError(f"No {self._pipeline} found in pipelines.")

        pipeline_name = key_parts[-1]
        registry = getattr(pipelines, self._pipeline)
        func = registry.get(pipeline_name, None)
        if func is None:
            raise ValueError(f"'{pipeline_name}' not found in {self._pipeline}.")

        # Compute, cache, return
        result = self._compute(instance, func)
        cache[cache_key] = result
        return result


def pipeline_cached_property(stage: ContextVar):
    """Decorator for context-sensitive cached properties backed by pipeline functions.

    The pipeline stage ordering is defined in ``PIPELINE_STAGES``. Passing a stage
    automatically includes all preceding stages in the cache key and resolves the
    pipeline attribute name.

    Args:
        stage: The ContextVar identifying the pipeline stage (e.g. ``NORMALIZER_NAME``).

    Example::

        @pipeline_cached_property(STANDARDIZER_NAME)
        def standardized(self, func):
            \"\"\"The standardized text.\"\"\"
            return func(self.raw)
    """
    context_vars, pipeline = _STAGE_INDEX[stage]

    def decorator(method: Callable) -> PipelineCachedProperty:
        return PipelineCachedProperty(
            context_vars=context_vars,
            pipeline=pipeline,
            compute=method,
            doc=method.__doc__,
        )

    return decorator



__all__ = [
    "display_basic_aligned",
    "ColorScheme",
    "DefaultColorScheme",
]


def __getattr__(name: str):
    if name in __all__:
        from bewer.reporting.python.alignment import (
            ColorScheme,
            DefaultColorScheme,
            display_basic_aligned,
        )

        _attrs = {
            "display_basic_aligned": display_basic_aligned,
            "ColorScheme": ColorScheme,
            "DefaultColorScheme": DefaultColorScheme,
        }
        return _attrs[name]
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



import string
from dataclasses import dataclass
from itertools import chain
from typing import TYPE_CHECKING

from error_align import error_align
from error_align.utils import OpType
from fuzzywuzzy import fuzz, process
from rapidfuzz.distance import Levenshtein

from bewer.core.text import _join_tokens
from bewer.metrics.base import METRIC_REGISTRY, ExampleMetric, Metric, MetricParams, metric_value
from bewer.metrics.cer import CER
from bewer.metrics.wer import WER

if TYPE_CHECKING:
    from bewer.core.example import Example

__all__ = [
    "LegacyKeywordAggregator",
    "LegacyMTR",
    "LegacyRMTR",
    "LegacyKeywordCER",
    "LegacyInsertions",
    "LegacyDeletionHallucinations",
    "LegacyHallucinationAggregator",
]

METRIC_REGISTRY.register_metric(
    WER,
    "legacy_wer",
    tokenizer="legacy",
    normalizer="legacy",
)
METRIC_REGISTRY.register_metric(
    WER,
    "legacy_wer_uncased",
    tokenizer="legacy",
    normalizer="legacy_uncased",
)
METRIC_REGISTRY.register_metric(
    WER,
    "legacy_wer_uncased_no_punct",
    tokenizer="legacy",
    normalizer="legacy_uncased_no_punct",
)

METRIC_REGISTRY.register_metric(
    CER,
    "legacy_cer",
    tokenizer="legacy",
    normalizer="legacy",
)
METRIC_REGISTRY.register_metric(
    CER,
    "legacy_cer_uncased",
    tokenizer="legacy",
    normalizer="legacy_uncased",
)
METRIC_REGISTRY.register_metric(
    CER,
    "legacy_cer_uncased_no_punct",
    tokenizer="legacy",
    normalizer="legacy_uncased_no_punct",
)


class _KeywordAggregator(ExampleMetric):
    @metric_value
    def cer_keyword(self) -> float:
        """Get the character error rate for medical terms."""
        return self._keyword_metrics["cer_keyword"]

    @metric_value
    def total_distance(self) -> float:
        """Get the total Levenshtein distance for medical terms."""
        return self._keyword_metrics["total_distance"]

    @metric_value
    def total_length(self) -> float:
        """Get the total length of medical terms."""
        return self._keyword_metrics["total_length"]

    @metric_value
    def match_count(self) -> int:
        """Get the number of exactly matched medical terms."""
        return self._keyword_metrics["match_count"]

    @metric_value
    def relaxed_match_count(self) -> int:
        """Get the number of medical terms matched with relaxed criteria."""
        return self._keyword_metrics["relaxed_match_count"]

    @metric_value
    def total_terms(self) -> int:
        """Get the total number of medical terms."""
        return self._keyword_metrics["total_terms"]

    @metric_value
    def correct_terms(self) -> list[str]:
        """Get the list of correctly matched medical terms."""
        return self._keyword_metrics["correct_terms"]

    @metric_value
    def _keyword_metrics(self) -> dict[str, float]:
        """Calculate keyword-focused metrics for medical terms."""

        total_distance = 0
        total_length = 0
        match_count = 0
        relaxed_match_count = 0
        correct_terms = []

        if "medical_terms" not in self.example.keywords:
            return dict(
                cer_keyword=0.0,
                total_distance=0.0,
                total_length=0.0,
                match_count=0,
                relaxed_match_count=0,
                total_terms=0,
                correct_terms=[],
            )

        ref = self.example.ref.joined(normalized=True)
        medical_terms = [kw for kw in self.example.keywords["medical_terms"] if kw.joined(normalized=True) in ref]
        max_n = max((len(term.tokens) for term in medical_terms), default=0)
        words = self.example.hyp.tokens.normalized
        ngram_matrix = [self._get_ngrams(words, n) for n in range(1, max_n + 1)]

        for term in medical_terms:
            term = _join_tokens(term.tokens, normalized=True)
            distance = self._term_distance(term, ngram_matrix=ngram_matrix)
            cer_score = distance / max(len(term), 1)  # Avoid division by zero
            if distance == 0:
                match_count += 1
                correct_terms.append(term)
            if cer_score <= self.params.cer_threshold:
                relaxed_match_count += 1
            total_distance += distance
            total_length += max(len(term), 1)

        cer_keyword = total_distance / total_length if total_length > 0 else 0

        return dict(
            cer_keyword=cer_keyword,
            total_distance=total_distance,
            total_length=total_length,
            match_count=match_count,
            relaxed_match_count=relaxed_match_count,
            total_terms=len(medical_terms),
            correct_terms=correct_terms,
        )

    @staticmethod
    def _get_ngrams(tokens: list[str], n: int) -> list[str]:
        """Get n-grams from a list of tokens."""
        if n < 1:
            raise ValueError("n must be a positive integer")
        if n == 1:
            return tokens
        return [" ".join(tokens[i : i + n]) for i in range(len(tokens) - n + 1)]

    @staticmethod
    def _term_distance(term: str, ngram_matrix: list[list[str]], retain_punc_list: list[str] = ["-", "/"]) -> int:
        """Calculate the Levenshtein distance between a term and a best match ngram."""
        # Select n-grams for the text
        n = len(term.split())
        ngrams = list(chain.from_iterable(ngram_matrix[:n]))

        # Find the best match using fuzzy matching
        best_match, _ = process.extractOne(term, ngrams, scorer=fuzz.ratio) if ngrams else ("", 0)

        # Define punctuation for the check as all punctuation except the ones we want to retain
        punct_to_remove = string.punctuation
        for p in retain_punc_list:
            punct_to_remove = punct_to_remove.replace(p, "")

        translator = str.maketrans("", "", punct_to_remove)
        punctuation_for_check = set(punct_to_remove)

        # Remove any punctuation (excluding retained punctuation)
        if not any(char in punctuation_for_check for char in term):
            best_match = best_match.translate(translator)

        best_match = best_match.lower()
        term = term.lower()

        # Calculate the Levenshtein distance
        return Levenshtein.distance(term, best_match)


@METRIC_REGISTRY.register("_legacy_kwa", tokenizer="legacy", normalizer="legacy_uncased")
class LegacyKeywordAggregator(Metric):
    short_name_base = "kwa"
    long_name_base = "Keyword Aggregator"
    description = "Aggregates keyword-focused metrics for medical terms."
    example_cls = _KeywordAggregator

    @dataclass
    class param_schema(MetricParams):
        cer_threshold: float = 0.2

    @metric_value
    def match_count(self) -> int:
        """Get the total number of exactly matched medical terms."""
        return sum([self.get_example_metric(example).match_count for example in self._src])

    @metric_value
    def relaxed_match_count(self) -> int:
        """Get the total number of medical terms matched with relaxed criteria."""
        return sum([self.get_example_metric(example).relaxed_match_count for example in self._src])

    @metric_value
    def total_terms(self) -> int:
        """Get the total number of medical terms."""
        return sum([self.get_example_metric(example).total_terms for example in self._src])

    @metric_value
    def total_length(self) -> float:
        """Get the total length of medical terms."""
        return sum([self.get_example_metric(example).total_length for example in self._src])

    @metric_value
    def total_distance(self) -> float:
        """Get the total Levenshtein distance of medical terms."""
        return sum([self.get_example_metric(example).total_distance for example in self._src])

    @metric_value
    def correct_terms(self) -> list[str]:
        """Get the list of correctly matched medical terms."""
        return list(chain.from_iterable([self.get_example_metric(example).correct_terms for example in self._src]))


@METRIC_REGISTRY.register("legacy_medical_word_accuracy")
class LegacyMTR(Metric):
    short_name_base = "MTR"
    long_name_base = "Medical Term Recall"
    description = (
        "Medical Term Recall (MTR) is computed as the number of correctly recognized medical terms "
        "divided by the total number of medical terms in the reference transcripts."
    )

    @metric_value(main=True)
    def value(self) -> float:
        """Get the medical term recall."""
        if self._src.metrics._legacy_kwa().total_terms == 0:
            return 1.0
        return self._src.metrics._legacy_kwa().match_count / self._src.metrics._legacy_kwa().total_terms


@METRIC_REGISTRY.register("legacy_relaxed_medical_word_accuracy")
class LegacyRMTR(Metric):
    short_name_base = "Relaxed MTR"
    long_name_base = "Relaxed Medical Term Recall"
    description = (
        "Relaxed Medical Term Recall (RMTR) is computed as the number of terms recognized with a relaxed criteria "
        "(i.e., within a certain character error rate threshold) divided by the total number of medical terms in the "
        "reference transcripts."
    )

    @metric_value(main=True)
    def value(self) -> float:
        """Get the medical term recall."""
        if self._src.metrics._legacy_kwa().total_terms == 0:
            return 1.0
        return self._src.metrics._legacy_kwa().relaxed_match_count / self._src.metrics._legacy_kwa().total_terms


@METRIC_REGISTRY.register("legacy_keyword_cer")
class LegacyKeywordCER(Metric):
    short_name_base = "Keyword CER"
    long_name_base = "Keyword Character Error Rate"
    description = (
        "Keyword Character Error Rate (Keyword CER) is computed as the character error rate for medical terms "
        "between the reference and hypothesis terms."
    )

    @metric_value(main=True)
    def value(self) -> float:
        """Get the medical character error rate."""
        if self._src.metrics._legacy_kwa().total_length == 0:
            return 1.0
        return self._src.metrics._legacy_kwa().total_distance / self._src.metrics._legacy_kwa().total_length


class _HallucinationAggregator(ExampleMetric):
    @metric_value
    def _insertion_metrics(self) -> int:
        """Get the number of hallucinated medical term insertions for this example."""
        ref = self.example.ref.joined(normalized=True)
        hyp = self.example.hyp.joined(normalized=True)
        alignments = error_align(ref=ref, hyp=hyp)

        # Check if the alignment has deletions
        consecutive_insertions = 0
        has_contiguous_insertions = False
        insertions = 0
        for alignment in alignments:
            if alignment.op_type == OpType.INSERT:
                consecutive_insertions += 1
                insertions += 1
                if consecutive_insertions > self.params.threshold:
                    has_contiguous_insertions = True
            else:
                consecutive_insertions = 0

        return dict(
            has_contiguous_insertions=has_contiguous_insertions,
            insertions=insertions,
        )

    @metric_value
    def has_contiguous_insertions(self) -> bool:
        """Check if there are contiguous hallucinated medical term insertions."""
        return self._insertion_metrics["has_contiguous_insertions"]

    @metric_value
    def insertions(self) -> int:
        """Get the number of hallucinated medical term insertions."""
        return self._insertion_metrics["insertions"]


@METRIC_REGISTRY.register("_legacy_hlcn")
class LegacyHallucinationAggregator(Metric):
    short_name_base = "Hallucination Insertions"
    long_name_base = "Hallucination Insertions"
    description = "Number of insertions that appear in "
    example_cls = _HallucinationAggregator

    @dataclass
    class param_schema(MetricParams):
        threshold: int = 2


@METRIC_REGISTRY.register("legacy_deletions")
class LegacyInsertions(Metric):
    short_name_base = "Hallucination Insertions"
    long_name_base = "Hallucination Insertions"
    description = "Average number of insertions per example."

    @metric_value(main=True)
    def value(self) -> int:
        """Get the number of hallucinated medical term insertions."""

        def get_insertions(example: "Example") -> int:
            return int(example.metrics._legacy_hlcn().insertions)

        return sum([get_insertions(example) for example in self._src]) / len(self._src)


@METRIC_REGISTRY.register("legacy_del_hallucinations")
class LegacyDeletionHallucinations(Metric):
    short_name_base = "Insertion Hallucinations"
    long_name_base = "Insertion Hallucinations"
    description = "Fraction of examples for which more than N consecutive insertions are observed."

    @metric_value(main=True)
    def value(self) -> int:
        """Get the number of examples with hallucinated deletions."""

        def get_int_value(example: "Example") -> int:
            return int(example.metrics._legacy_hlcn().has_contiguous_insertions)

        return sum([get_int_value(example) for example in self._src]) / len(self._src)



from contextlib import contextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING, Optional

from bewer.flags import DEFAULT

__all__ = ["set_pipeline", "get_standardizer", "get_tokenizer", "get_normalizer"]

if TYPE_CHECKING:
    from bewer import Dataset
    from bewer.preprocessing.normalization import Normalizer
    from bewer.preprocessing.tokenization import Tokenizer

# Define the context variables with default value "default"
STANDARDIZER_NAME = ContextVar("STANDARDIZER", default=DEFAULT)
TOKENIZER_NAME = ContextVar("TOKENIZER", default=DEFAULT)
NORMALIZER_NAME = ContextVar("NORMALIZER", default=DEFAULT)

# Ordered pipeline stages: (context_var, pipelines attribute name)
PIPELINE_STAGES = (
    (STANDARDIZER_NAME, "standardizers"),
    (TOKENIZER_NAME, "tokenizers"),
    (NORMALIZER_NAME, "normalizers"),
)


@contextmanager
def set_pipeline(standardizer=DEFAULT, tokenizer=DEFAULT, normalizer=DEFAULT):
    tok_token = TOKENIZER_NAME.set(tokenizer)
    std_token = STANDARDIZER_NAME.set(standardizer)
    norm_token = NORMALIZER_NAME.set(normalizer)
    try:
        yield
    finally:
        TOKENIZER_NAME.reset(tok_token)
        STANDARDIZER_NAME.reset(std_token)
        NORMALIZER_NAME.reset(norm_token)


def get_standardizer(dataset: "Dataset") -> Optional["Normalizer"]:
    """Get the standardizer for the dataset in the current context."""
    return dataset.pipelines.standardizers.get(STANDARDIZER_NAME.get(), None)


def get_tokenizer(dataset: "Dataset") -> Optional["Tokenizer"]:
    """Get the tokenizer for the dataset in the current context."""
    return dataset.pipelines.tokenizers.get(TOKENIZER_NAME.get(), None)


def get_normalizer(dataset: "Dataset") -> Optional["Normalizer"]:
    """Get the normalizer for the dataset in the current context."""
    return dataset.pipelines.normalizers.get(NORMALIZER_NAME.get(), None)



from io import StringIO
from typing import TYPE_CHECKING

from rich.console import Console
from rich.text import Text

if TYPE_CHECKING:
    from bewer.core.token import TokenList

__all__: list[str] = []  # Internal


def highlight_span(text: str, start: int, end: int, style: str) -> str:
    # Use a StringIO buffer to capture ANSI output
    buffer = StringIO()
    console = Console(file=buffer, force_terminal=True, color_system="truecolor")

    # Create styled text
    text = Text(text)
    text.stylize(style, start, end)

    # Print to the buffer instead of the terminal
    console.print(text, end="")

    # Get the ANSI string
    return buffer.getvalue()


def highlight_tokens(text: str, tokens: "TokenList", style1: str = "green", style2: str = "yellow") -> str:
    # Use a StringIO buffer to capture ANSI output
    buffer = StringIO()
    console = Console(file=buffer, force_terminal=True, color_system="truecolor")

    # Create styled text
    text = Text(text)
    for token in tokens:
        text.stylize(style1, token.start, token.end)
        style1, style2 = style2, style1

    # Print to the buffer instead of the terminal
    console.print(text, end="")

    # Get the ANSI string
    return buffer.getvalue()



from enum import Enum
from functools import cached_property
from typing import TYPE_CHECKING, Iterable, Optional, Union

import regex as re

from bewer.core.caching import pipeline_cached_property
from bewer.core.token import Token
from bewer.preprocessing.context import NORMALIZER_NAME, STANDARDIZER_NAME, TOKENIZER_NAME

if TYPE_CHECKING:
    from bewer.core.example import Example

__all__ = ["Text", "TextType", "TokenList"]


class TextType(str, Enum):
    REF = "ref"
    HYP = "hyp"
    KEYWORD = "keyword"


def _join_tokens(tokens: "TokenList", normalized: bool = True) -> str:
    """Join tokens into a single string, preserving original spacing.

    Args:
        tokens (TokenList): The tokens to join.
        normalized (bool): Whether to use normalized token text.

    Returns:
        str: The joined string.
    """
    joined = ""
    prev_end = 0
    for token in tokens:
        if token.start > prev_end:
            joined += f" {token.normalized}" if normalized else f" {token.raw}"
        else:
            joined += token.normalized if normalized else token.raw
        prev_end = token.end
    return joined.strip()


class Text:
    """BeWER text representation.

    Attributes:
        raw (str): The original text string.
        standardized (str): The standardized text string after applying the specified standardizer.
        src (Example): The source Example object that this text belongs to.
        text_type (TextType): The type of the text (reference, hypothesis, or keyword).
    """

    def __init__(
        self,
        raw: str | None,
        src: Optional["Example"] = None,
        text_type: Optional[TextType] = None,
    ):
        """Initialize the Text object.

        Args:
            raw: The original text string (reference, hypothesis, or keyword).
            src: Parent Example object. Can be set later via set_source().
            text_type: The type of the text (REF, HYP, or KEYWORD).
        """
        self._raw = raw
        self._text_type = text_type

        self._cache_standardized = {}
        self._cache_tokens = {}

        self._src = None
        self._pipelines = None
        if src is not None:
            self.set_source(src)

    @property
    def raw(self) -> str:
        if self._raw is None:
            raise ValueError("Raw text is None and lazy inference is not implemented.")
        return self._raw

    @property
    def src(self) -> Optional["Example"]:
        """Get the parent Example object."""
        return self._src

    @property
    def text_type(self) -> Optional[TextType]:
        return self._text_type

    @property
    def pipelines(self):
        return self._pipelines

    @pipeline_cached_property(STANDARDIZER_NAME)
    def standardized(self, standardizer) -> str:
        """The standardized text string after applying the active standardizer."""
        return standardizer(self.raw)

    @pipeline_cached_property(TOKENIZER_NAME)
    def tokens(self, tokenizer) -> "TokenList":
        """The list of Token objects produced by the active tokenizer."""
        return TokenList.from_matches(tokenizer(self.standardized), src=self)

    def set_source(self, src: "Example") -> None:
        """Set the parent Example object.

        Args:
            src: The parent Example object.

        Raises:
            ValueError: If source is already set.
        """
        if self._src is not None:
            raise ValueError("Source already set for Text")

        self._src = src

        # Cache pipeline reference
        self._pipelines = src.pipelines if src is not None else None

    def joined(self, normalized: bool = True) -> str:
        """Get the joined text from tokens.

        Args:
            normalized (bool): Whether to use normalized tokens.
        """
        return _join_tokens(self.tokens, normalized=normalized)

    def __hash__(self):
        return hash((self.raw, self._text_type))

    def __repr__(self):
        text = self.raw if len(self.raw) <= 46 else self.raw[:46] + "..."
        return f'Text("{text}")'


class TokenList(tuple["Token", ...]):
    """An immutable sequence of Token objects."""

    def __new__(cls, iterable=(), src=None):
        return super().__new__(cls, iterable)

    def __init__(self, iterable=(), src: Optional["Text"] = None):
        self._normalized_index_cache: dict[str, dict[str, set[int]]] = {}
        self._normalized_cache: dict[str, list[str]] = {}
        self._src = src

    @property
    def src(self) -> Optional["Text"]:
        """Get the source Text object."""
        return self._src

    @classmethod
    def from_matches(
        cls,
        matches: "Iterable[re.Match]",
        src: Optional["Text"] = None,
    ) -> "TokenList":
        """Create a TokenList from an iterable of regex match objects.

        Args:
            matches: An iterable of regex Match objects.
            src: The source Text object, if available.

        Returns:
            TokenList: A list of Token objects created from the matches.
        """
        return cls((Token.from_match(match, index=i, src=src) for i, match in enumerate(matches)), src=src)

    @property
    def raw(self) -> list[str]:
        """Get the raw tokens as a regular Python list.

        Returns:
            list[str]: The raw tokens.
        """
        return [token.raw for token in self]

    @property
    def normalized(self) -> list[str]:
        """Get the normalized tokens as a regular Python list.

        Returns:
            list[str]: The normalized tokens.
        """
        key = NORMALIZER_NAME.get()
        if key not in self._normalized_cache:
            self._normalized_cache[key] = [token.normalized for token in self]
        return self._normalized_cache[key]

    def ngrams(
        self,
        n: int,
        normalized: bool = True,
        join_tokens: bool = True,
    ) -> list[str]:
        """Get n-grams from the token list.

        Args:
            n (int): The size of the n-grams.
            normalized (bool): Whether to use normalized tokens.
            join_tokens (bool): Whether to join tokens into a single string.

        Returns:
            list[str]: The list of n-grams.
        """
        if n < 1:
            raise ValueError("n must be a positive integer")
        if n == 1:
            return self.raw if not normalized else self.normalized
        ngrams = []
        for i in range(len(self) - n + 1):
            ngram = self[i : i + n]
            if join_tokens:
                ngram = _join_tokens(ngram, normalized=normalized)
            else:
                ngram = ngram.normalized if normalized else ngram.raw
            ngrams.append(ngram)
        return ngrams

    @cached_property
    def _raw_index_mapping(self) -> dict[str, set[int]]:
        """Mapping from raw token text to set of positions in this TokenList."""
        mapping: dict[str, set[int]] = {}
        for i, token in enumerate(self):
            mapping.setdefault(token.raw, set()).add(i)
        return mapping

    @property
    def _normalized_index_mapping(self) -> dict[str, set[int]]:
        """Mapping from normalized token text to set of positions, cached per normalizer."""
        key = NORMALIZER_NAME.get()
        cache = self._normalized_index_cache
        if key not in cache:
            mapping: dict[str, set[int]] = {}
            for i, token in enumerate(self):
                mapping.setdefault(token.normalized, set()).add(i)
            cache[key] = mapping
        return cache[key]

    def indices(self, text: str, normalized: bool = True) -> set[int]:
        """Find all token positions where the token text matches the given string.

        Args:
            text: The string to search for.
            normalized: If True, compare against normalized token text.
                        If False, compare against raw token text.

        Returns:
            set[int]: Set of indices where the token's text matches.
        """
        mapping = self._normalized_index_mapping if normalized else self._raw_index_mapping
        return mapping.get(text, set())

    def _sub_repr(self):
        """Used internally by TextTokenList.__repr__"""
        tokens = self[:5]
        tokens_str = ",  ".join([repr(token) for token in tokens])
        if len(self) > 5:
            tokens_str += ", ..."
        return f"TokenList([{tokens_str}])"

    def __getitem__(self, index: int | slice) -> Union["Token", "TokenList"]:
        if isinstance(index, slice):
            return TokenList(super().__getitem__(index))
        return super().__getitem__(index)

    def __add__(self, other: "TokenList") -> "TokenList":
        return TokenList(super().__add__(other))

    def __repr__(self):
        tokens = self[:60]
        tokens_str = ",\n ".join([repr(token) for token in tokens])
        if len(self) > 60:
            tokens_str += ",\n ..."
        return f"TokenList([\n {tokens_str}]\n)"
